#!/usr/bin/env python
"""idifference.

DEVELOPMENT NOTE: This implementation will soon be replaced by what is currently idifference2.py, after a period of testing by users.  If idifference2.py does not meet your needs, but idifference.py does, please let one of the maintainers know (email addresses in the Git history or the python/ChangeLog file).

Generates a report about what's different between two disk images.

Process:

1. A map is kept of all filenames->sha1 codes and all sha1->filenames.
2. For each image, read all of the fileobject objects:
    - create new maps
    - Note when things change.
    - Delete each file in the old map as it is processed.
3. Report files left in map; that's the files that were deleted!
4. Replace the old maps with the new maps
"""

__version__ = "0.2.0rfc4"

import sys,fiwalk,dfxml,time
import copy
import logging
if sys.version_info < (3,1):
    raise RuntimeError("idifference.py now requires Python 3.1 or above")

#Global variable, to be adjusted later
options = None

def ignore_filename(fn, include_dotdirs=False):
    """
    Ignores particular file name patterns output by TSK.  Detecting new
    and renamed files becomes difficult if there are 3+ names for an
    inode (i.e. "dir", "dir/.", and "dir/child/..").
    """
    return (not include_dotdirs and (fn.endswith("/.") or fn.endswith("/.."))) or fn in set(['$FAT1','$FAT2'])    

def ptime(t):
    """Print the time in the requested format. T is a dfxml time value.  If T is null, return 'null'."""
    global options
    if t is None:
        return "null"
    if options and options.timestamp:
        return str(t.timestamp())
    else:
        return str(t.iso8601())

def dprint(x):
    global options
    if options and options.debug: print(x)

def header():
    global options
    if options and options.html:
        print("""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN http://www.w3.org/TR/html4/loose.dtd">
<html>
<body>
<style>
body  { font-family: Sans-serif;}
.sha1 { font-family: monospace; font-size: small;}
.filesize { padding-left: 15px; padding-right: 15px; text-align: right;}
</style>
""")

def h1(title):
    global options
    if options and options.html:
        print("<h1>%s</h1>" % title)
        return
    print("\n\n%s\n" % title)

def h2(title):
    global options
    if options and options.html:
        print("<h2>%s</h2>" % title)
        return
    print("\n\n%s\n%s" % (title,"="*len(title)))


def table(rows,styles=None,break_on_change=False):
    import sys
    global options
    def alldigits(x):
        if type(x)!=str: return False
        for ch in x:
            if ch.isdigit()==False: return False
        return True

    def fmt(x):
        if x==None: return ""
        if type(x)==int or alldigits(x):
            return "{0:>12}".format(x)
        return str(x)
            
    if options and options.html:
        print("<table>")
        for row in rows:
            print("<tr>")
            if not styles:
                styles = [""]*len(rows)
            for (col,style) in zip(row,styles):
                sys.stdout.write("<td class='%s'>%s</td>" % (style,col))
            print("<tr>")
        print("</table>")
        return
    lastRowCol0 = None
    for row in rows:
        # Insert a blank line if this row[0] is not the same as last row[0]
        if row[0]!=lastRowCol0 and break_on_change:
            sys.stdout.write("\n")
            lastRowCol0 = row[0]
        # Write the row.
        # This won't generate a unicode encoding error because the strings are valid unicode.
        sys.stdout.write("\t".join([fmt(col) for col in row]))
        sys.stdout.write("\n")

#
# This program keeps track of the current and previous diskstate in a single
# object called "DiskState". Another way to do that would have been to have
# the instance built from the XML file and then have another function that compares
# them.
#        

class DiskState:
    global options

    def __init__(self,notimeline=False,summary=False,include_dotdirs=False):
        self.current_fname = None # This class field is the name of the current disk image, whereas other fnames are in-image file names
        self.new_fnames = dict() # maps from fname -> fi
        self.new_inodes = dict() # maps from (partition, inode_number) -> fi
        self.new_fi_tally = 0
        self.notimeline = notimeline
        self.summary = summary
        self.include_dotdirs = include_dotdirs
        self.changed_mtime_tally = 0
        self.changed_atime_tally = 0
        self.changed_ctime_tally = 0
        self.changed_crtime_tally = 0
        self.changed_dir_sha1_tally = 0
        self.changed_file_sha1_tally = 0
        self.changed_filesize_tally = 0
        self.changed_first_byterun_tally = 0
        self.next()
        
    def next(self):
        """Called when the next image is processed."""
        global options
        self.fnames = self.new_fnames
        self.inodes = self.new_inodes
        self.fi_tally = self.new_fi_tally
        self.new_fnames = dict()
        self.new_inodes = dict()
        #Reset sets
        self.new_files          = set()     # set of file objects
        self.renamed_files      = set()     # set of (oldfile,newfile) file objects
        self.changed_content    = set()     # set of (oldfile,newfile) file objects
        self.changed_properties = set()     # list of (oldfile,newfile) file objects
        #Reset counters
        self.new_fi_tally = 0
        if self.notimeline:
            self.timeline = None
        else:
            self.timeline = set()
        self.changed_mtime_tally = 0
        self.changed_atime_tally = 0
        self.changed_ctime_tally = 0
        self.changed_crtime_tally = 0
        self.changed_dir_sha1_tally = 0
        self.changed_file_sha1_tally = 0
        self.changed_filesize_tally = 0
        self.changed_first_byterun_tally = 0

    def process_fi(self,fi):
        global options
        # Filter out specific filenames create by TSK that are not of use
        if ignore_filename(fi.filename(), self.include_dotdirs):
            return 

        dprint("processing %s" % str(fi))
        
        # See if the filename changed its hash code
        changed = False
        if not fi.allocated():
            return # only look at allocated files

        # Remember the file for the next generation
        self.new_fnames[fi.filename()] = fi
        self.new_inodes[(fi.partition(), fi.inode())] = fi
        self.new_fi_tally += 1

        # See if a file with this filename had its contents change or properties changed
        ofi = self.fnames.get(fi.filename(),None)
        if ofi:
            dprint("   found ofi")
            any_diff = False
            if ofi.sha1()!=fi.sha1():
                dprint("      >>> sha1 changed")
                self.changed_content.add((ofi,fi))
                any_diff = True
            elif ofi.atime() != fi.atime() or \
                    ofi.mtime() != fi.mtime() or \
                    ofi.crtime() != fi.crtime() or \
                    ofi.ctime() != fi.ctime():
                dprint("      >>> time changed")
                self.changed_properties.add((ofi,fi))
                any_diff = True

            if any_diff:
                #Count the types of changes that happened
                if ofi.filesize() != fi.filesize():
                    self.changed_filesize_tally += 1
                if ofi.sha1() != fi.sha1():
                    if ofi.is_dir():
                        self.changed_dir_sha1_tally += 1
                    elif ofi.is_file():
                        self.changed_file_sha1_tally += 1
                if ofi.mtime() != fi.mtime():
                    self.changed_mtime_tally += 1
                if ofi.atime() != fi.atime():
                    self.changed_atime_tally += 1
                if ofi.ctime() != fi.ctime():
                    self.changed_ctime_tally += 1
                if ofi.crtime() != fi.crtime():
                    self.changed_crtime_tally += 1
                if ofi.byte_runs() and fi.byte_runs():
                    brdiff = 0
                    ofirstbr = ofi.byte_runs()[0]
                    nfirstbr =  fi.byte_runs()[0]
                    try:
                        if ofirstbr.file_offset == nfirstbr.file_offset:
                            brdiff = 1
                        if ofirstbr.img_offset == nfirstbr.img_offset:
                            brdiff = 1
                        if ofirstbr.fs_offset == nfirstbr.fs_offset:
                            brdiff = 1
                    except:
                        pass
                    self.changed_first_byterun_tally += brdiff
          

        # If a new file, note that (and optionally add to the timeline)
        if not ofi:
            self.new_files.add(fi)
            if self.timeline:
                create_time = fi.crtime()
                if not create_time: create_time = fi.ctime()
                if not create_time: create_time = fi.mtime()
                self.timeline.add((create_time,fi.filename(),"created"))

        # Delete files we have seen (so we can find out the files that were deleted)
        if fi.filename() in self.fnames:
            del self.fnames[fi.filename()]

        # Look for files that were renamed
        ofi = self.inodes.get((fi.partition(), fi.inode()),None)
        if ofi and ofi.filename() != fi.filename() and ofi.sha1()==fi.sha1():
            #Never consider current-directory or parent-directory for rename operations.  Because we match on partition+inode numbers, these trivially match.
            if not (fi.filename().endswith("/.") or fi.filename().endswith("/..") or ofi.filename().endswith("/.") or ofi.filename().endswith("/..")):
                self.renamed_files.add((ofi,fi))

    def process(self,fname):
        self.prior_fname = self.current_fname
        self.current_fname = fname
        if fname.endswith("xml"):
            with open(fname,'rb') as xmlfile:
                for fi in dfxml.iter_dfxml(xmlfile, preserve_elements=True):
                    self.process_fi(fi)
        else:
            fiwalk.fiwalk_using_sax(imagefile=open(fname,'rb'), flags=fiwalk.ALLOC_ONLY, callback=self.process_fi)

    def print_fis(self,title,fis):
        h2(title)
        def fidate(fi):
            try:
                return str(ptime(fi.mtime()))
            except TypeError:
                return "n/a"
        res = [(fidate(fi),str(fi.filesize()),fi.filename()) for fi in fis]
        if res:
            table(sorted(res))

    def print_fi2(self,title,fi2s):
        def prtime(t):
            return "%d (%s)" % (t,ptime(t))

        h2(title)
        res = set()
        for(ofi,fi) in fi2s:
            if ofi.filename() != fi.filename():
                res.add((ofi.filename(),"renamed to",fi.filename()))
                # Don't know when it was renamed
            if ofi.filesize() != fi.filesize():
                res.add((ofi.filename(),"resized",ofi.filesize(),"->",fi.filesize()))
                if self.timeline: self.timeline.add((fi.mtime(),fi.filename(),"resized",ofi.filesize(),"->",fi.filesize()))
            if ofi.sha1() != fi.sha1():
                res.add((ofi.filename(),"SHA1 changed",ofi.sha1(),"->",fi.sha1()))
                if self.timeline: self.timeline.add((fi.mtime(),fi.filename(),"SHA1 changed",ofi.sha1(),"->",fi.sha1()))
            if ofi.atime() != fi.atime():
                if not options.noatime:
                    res.add((ofi.filename(),"atime changed",ptime(ofi.atime()),"->",ptime(fi.atime())))
                    if self.timeline: self.timeline.add((fi.atime(),fi.filename(),"atime changed",prtime(ofi.atime()),"->",prtime(fi.atime())))
            if ofi.mtime() != fi.mtime():
                res.add((ofi.filename(),"mtime changed",ptime(ofi.mtime()),"->",ptime(fi.mtime())))
                if self.timeline: self.timeline.add((fi.mtime(),fi.filename(),"mtime changed",prtime(ofi.mtime()),"->",prtime(fi.mtime())))
            if ofi.ctime() != fi.ctime():
                res.add((ofi.filename(),"ctime changed",ptime(ofi.ctime()),"->",ptime(fi.ctime())))
                if self.timeline: self.timeline.add((fi.ctime(),fi.filename(),"ctime changed",prtime(ofi.ctime()),"->",prtime(fi.ctime())))
            if ofi.crtime() != fi.crtime():
                res.add((ofi.filename(),"crtime changed",ptime(ofi.crtime()),"->",ptime(fi.crtime())))
                if self.timeline: self.timeline.add((fi.crtime(),fi.filename(),"crtime changed",prtime(ofi.crtime()),"->",prtime(fi.crtime())))
        if res:
            table(sorted(res),break_on_change=True)

    def print_timeline(self):
        prt = []

        # Make the dates printable
        for line in sorted(self.timeline):
            prt.append([ptime(line[0])]+list(line[1:]))
        h2("Timeline")
        table(prt)

    def to_xml(self):
        import xml.etree.ElementTree as ET
        
        ET.register_namespace("delta", dfxml.XMLNS_DELTA)
        
        if not options.xmlfilename:
            sys.stderr.write("XML output filename not specified.\n")
            exit(1)

        metadict = dict()
        metadict["XMLNS_DFXML"] = dfxml.XMLNS_DFXML
        metadict["XMLNS_DELTA"] = dfxml.XMLNS_DELTA
        metadict["program"] = sys.argv[0]
        metadict["version"] = __version__
        metadict["commandline"] = " ".join(sys.argv)
        metadict["priorf"] = self.prior_fname
        metadict["currentf"] = self.current_fname

        xmlfile = open(options.xmlfilename, "w")
        xmlfile.write("""\
<?xml version="1.0" encoding="UTF-8"?>
<dfxml
  version="1.0"
  xmlns='%(XMLNS_DFXML)s'
  xmlns:dc='http://purl.org/dc/elements/1.1/'
  xmlns:delta='%(XMLNS_DELTA)s'
  xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'>
  <metadata>
    <dc:type>Disk Image Difference Manifest</dc:type>
  </metadata>
  <creator>
    <program>%(program)s</program>
    <version>%(version)s</version>
    <execution_environment>
      <command_line>%(commandline)s</command_line>
    </execution_environment>
  </creator>
  <source>
    <image_filename>%(priorf)s</image_filename>
    <image_filename>%(currentf)s</image_filename>
  </source>
""" % metadict)

        def _annotate_changes(tmpel, ofi, fi):
            """
            Adds "delta:changed_property" attributes to elements that changed their values.
            Returns number of annotations needed.
            """
            retval = 0
            def _xpaths(xp):
                """
                Returns a list of xpaths: First, with an xmlns; second, as input.

                @param xp
                An xpath expression where all elements and attributes needing a namespace declaration are prefixed with "{0}" (for Python string formatting).
                """
                retval = []
                for nsprefix in ["{" + dfxml.XMLNS_DFXML + "}", ""]:
                    retval.append(xp.format(nsprefix))
                return retval
            # Triplets: Old value, new value, XPaths to find element to annotate
            for (oval, nval, xpaths) in [
              (ofi.filename(), fi.filename(), _xpaths("./{0}filename")),
              (ofi.sha1(), fi.sha1(), _xpaths("./{0}hashdigest[@type='sha1']")),
              (ofi.md5(), fi.md5(), _xpaths("./{0}hashdigest[@type='md5']")),
              (ofi.mtime(), fi.mtime(), _xpaths("./{0}mtime")),
              (ofi.atime(), fi.atime(), _xpaths("./{0}atime")),
              (ofi.ctime(), fi.ctime(), _xpaths("./{0}ctime")),
              (ofi.crtime(), fi.crtime(), _xpaths("./{0}crtime")),
              (ofi.filesize(), fi.filesize(), _xpaths("./{0}filesize"))
            ]:
                #Find and flag the changed properties

                #Skip null-null comparisons
                if oval is None and nval is None:
                    continue

                if oval != nval:
                    retval += 1
                    #Find first namespace match for the property element 
                    for xp in xpaths:
                        propertyel = tmpel.find(xp)
                        if not propertyel is None:
                            break
                    if propertyel is None:
                        comment = ET.Comment("Warning: Tried to note a changed property with the XPath queries %r; however, could not find the element." % xpaths)
                        tmpel.insert(0, comment)
                    else:
                        propertyel.attrib["delta:changed_property"] = "1"
            return retval

        #List new files
        for fi in self.new_files:
            #xmlfile.write("  <!-- + %s -->\n" % fi.filename())
            xmlfile.write("  ")
            tmpel = copy.copy(fi.xml_element)
            tmpel.attrib["delta:new_file"] = "1"
            xmlfile.write(dfxml.ET_tostring(tmpel, encoding="unicode"))
            xmlfile.write("\n")
        #List deleted files
        for fi in self.fnames.values():
            #xmlfile.write("<!-- - %s -->\n" % fi.filename())
            xmlfile.write("  ")
            tmpel = ET.Element("fileobject")
            tmpel.attrib["delta:deleted_file"] = "1"
            tmpchild = copy.copy(fi.xml_element)
            tmpchild.tag = "delta:original_fileobject"
            tmpel.insert(-1, tmpchild)
            xmlfile.write(dfxml.ET_tostring(tmpel, encoding="unicode"))
            xmlfile.write("\n")
        #List renamed files
        for (ofi, fi) in self.renamed_files:
            #xmlfile.write("<!-- ! %s -> %s -->\n" % (ofi.filename(), fi.filename()))
            tmpel = copy.copy(fi.xml_element)
            annos = _annotate_changes(tmpel, ofi, fi)
            tmpoldel = copy.copy(ofi.xml_element)
            tmpoldel.tag = "delta:original_fileobject"
            tmpel.append(tmpoldel)
            tmpel.attrib["delta:renamed_file"] = "1"
            if annos > 1:
                tmpel.attrib["delta:changed_file"] = "1"
            xmlfile.write(dfxml.ET_tostring(tmpel, encoding="unicode"))
            xmlfile.write("\n")
        #List files with with modified data or metadata
        changed_files = set.union(set(self.changed_content), set(self.changed_properties))
        for (ofi, fi) in changed_files:
            #xmlfile.write("<!-- ~ %s -->\n" % fi.filename())
            xmlfile.write("  ")
            tmpel = copy.copy(fi.xml_element)
            _annotate_changes(tmpel, ofi, fi)
            tmpoldel = copy.copy(ofi.xml_element)
            tmpoldel.tag = "delta:original_fileobject"
            tmpel.append(tmpoldel)
            tmpel.attrib["delta:changed_file"] = "1"
            xmlfile.write(dfxml.ET_tostring(tmpel, encoding="unicode"))
            xmlfile.write("\n")
        xmlfile.write("</dfxml>\n")
        xmlfile.close()

    def report(self):
        header()
        h1("Disk image:"+self.current_fname)
        self.print_fis("New Files:",self.new_files)
        self.print_fis("Deleted Files:",self.fnames.values())
        self.print_fi2("Renamed Files:",self.renamed_files)
        self.print_fi2("Files with modified content:",self.changed_content)
        self.print_fi2("Files with changed file properties:",self.changed_properties)
        if self.summary:
            h2("Summary:")
            table([
              ("Prior image's file (file object) tally", str(self.fi_tally)),
              ("Prior image's file (inode) tally", str(len(self.inodes))),
              ("Current image's file (file object) tally", str(self.new_fi_tally)),
              ("Current image's file (inode) tally", str(len(self.new_inodes))),
              ("New files", str(len(self.new_files))),
              ("Deleted files", str(len(self.fnames))),
              ("Renamed files", str(len(self.renamed_files))),
              ("Files with modified content", str(len(self.changed_content))),
              ("Files with changed file properties", str(len(self.changed_properties)))
            ])

        if self.timeline: self.print_timeline()

    def output_archive(self,imagefile=None,tarname=None,zipname=None):
        """Write the changed and/or new files to a tarfile or a ZIP file. """
        import zipfile, tarfile, StringIO, datetime

        tfile = None
        zfile = None

        to_archive = self.new_files.copy()
        to_archive = to_archive.union(set([val[1] for val in self.changed_content]))
        to_archive = to_archive.union(set([val[1] for val in self.changed_properties]))

        # Make sure we are just writing out inodes that have file contents
        to_archive = filter(
            lambda fi:fi.allocated() and fi.has_tag("inode") and fi.has_contents()
            and (fi.name_type()=='' or fi.name_type()=='r'),
            to_archive)

        if len(to_archive)==0:
            print("No archive created, as no allocated files created or modified")
            return

        if tarname:
            print(">>> Creating tar file: %s" % tarname)
            tfile = tarfile.TarFile(tarname,mode="w")

        if zipname:
            print(">>> Creating zip file: %s" % zipname)
            zfile = zipfile.ZipFile(zipname,mode="w",allowZip64=True)

        files_written=set()
        content_error_log = []
        for fi in to_archive:
            filename = fi.filename()
            fncount = 1
            while filename in files_written:
                filename = "%s.%d" % (fi.filename(),fnperm)
                fncount+= 1
            contents = None
            try:
                contents = fi.contents(imagefile)
            except ValueError as ve:
                if ve.message.startswith("icat error"):
                    #Some files cannot be recovered, even from images that do not seem corrupted; log the icat command that failed.
                    content_error_log.append(ve.message)
                else:
                    #This is a more interesting error, so have process die to report immediately.
                    raise
            if contents:
                if tfile:
                    info = tarfile.TarInfo(name=filename)
                    info.mtime = fi.mtime()
                    info.atime = fi.atime()
                    info.ctime = fi.ctime()
                    info.uid   = fi.uid()
                    info.gid   = fi.gid()
                    info.size  = fi.filesize()
                    # addfile requires a 'file', so let's make one
                    string = StringIO.StringIO()
                    string.write(contents)
                    string.seek(0)
                    tfile.addfile(tarinfo=info,fileobj=string)
                if zfile:
                    mtimestamp = fi.mtime().timestamp()
                    info = zipfile.ZipInfo(filename)
                    if mtimestamp:
                        #mtime might be null
                        info.date_time = datetime.datetime.fromtimestamp(mtimestamp).utctimetuple()
                    info.internal_attr = 1
                    info.external_attr = 2175008768 # specifies mode 0644
                    zfile.writestr(info,contents)
        if tfile: tfile.close()
        if zfile: zfile.close()
        if len(content_error_log) > 0:
            sys.stderr.write("Errors retrieving file contents:\n")
            sys.stderr.write("\n".join(content_error_log))
            sys.stderr.write("\n")

if __name__=="__main__":
    from optparse import OptionParser
    global options

    parser = OptionParser()
    parser.usage = '%prog [options] file1 file2 [file3...]  (files can be xml or image files)'
    parser.add_option("-x","--xml",help="specify output file for XML",dest="xmlfilename")
    parser.add_option("--html",help="specify output in HTML",action="store_true")
    parser.add_option("-n","--notimeline",help="do not generate a timeline",action="store_true")
    parser.add_option("-d","--debug",help="debug",action='store_true')
    parser.add_option("-T","--tararchive",help="create tar archive file of new/changed files",dest="tarfile")
    parser.add_option("-Z","--zipfile",help="create ZIP64 archive file of new/changed files",dest="zipfile")
    parser.add_option("--include-dotdirs",help="include files with names ending in '/.' and '/..'",action="store_true", dest="include_dotdirs", default=False)
    parser.add_option("--summary",help="output summary statistics of file system changes",action="store_true", default=False)
    parser.add_option("--timestamp",help="output all times in Unix timestamp format; otherwise use ISO 8601",action="store_true")
    parser.add_option("--imagefile",help="specifies imagefile or file2 is an XML file and you are archiving")
    parser.add_option("--noatime",help="Do not include atime changes",action="store_true")

    (options,args) = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if options.debug else logging.INFO)

    if len(args)<1:
        parser.print_help()
        sys.exit(1)

    if (options.tarfile or options.zipfile) and args[1].endswith("xml") and not options.imagefile:
        print("ERROR: %s is NOT an XML file and no imagefile specified" % args[1])
        parser.print_help()
        exit(1)

    s = DiskState(notimeline=options.notimeline, summary=options.summary, include_dotdirs=options.include_dotdirs)
    for infile in args:
        print(">>> Reading %s" % infile)
        s.process(infile)
        if infile!=args[0]:
            # Not the first file. Report and optionally archive
            if options.tarfile or options.zipfile:
                imagefilename = infile
                if imagefilename.endswith("xml"):
                    imagefilename = options.imagefile
                s.output_archive(imagefile=open(imagefilename),tarname=options.tarfile,zipname=options.zipfile)
            if options.xmlfilename:
                s.to_xml()
            else:
                s.report()
        s.next()




__version__ = "0.3.0"

import Objects
import logging
import os
import hashlib
import sqlite3

_logger = logging.getLogger(os.path.basename(__file__))

_nagged_ids = False
_used_ids = set()
_last_id = 1

def _generate_id():
    """
    Creates an ID number unique to all DFXML documents ran through write_sector_hashes_to_db in this process.
    """
    global _used_ids
    global _last_id
    while _last_id in _used_ids:
        _last_id += 1
    _used_ids.add(_last_id)
    return _last_id

sql_schema_files = """CREATE TABLE files(
  obj_id INTEGER NOT NULL,
  partition INTEGER,
  inode INTEGER,
  filename TEXT,
  filesize INTEGER
);"""
sql_schema_block_hashes = """CREATE TABLE block_hashes(
  obj_id INTEGER NOT NULL,
  img_offset INTEGER,
  fs_offset INTEGER,
  file_offset INTEGER,
  len INTEGER NOT NULL,
  md5 TEXT,
  sha1 TEXT
);"""

def write_sector_hashes_to_db(raw_image, dfxml_doc, predicate, db_output_path, pad_sectors=False):
    """
    Produces sector hashes of all files that fit a predicate.
    Predicate function: Takes a FileObject as input; returns True if the FileObject should have its sectors hashed (if possible).
    """
    global _used_ids

    if os.path.exists(db_output_path):
        raise ValueError("Database output path exists.  Aborting - will not overwrite.  (Path: %r.)" % db_output_path)

    conn = sqlite3.connect(db_output_path)
    conn.isolation_level = "EXCLUSIVE"
    conn.row_factory = sqlite3.Row
    cursor = conn.cursor()

    cursor.execute(sql_schema_files)
    cursor.execute(sql_schema_block_hashes)
    conn.commit()

    for (obj_no, obj) in enumerate(dfxml_doc):
        if not isinstance(obj, Objects.FileObject):
            continue
        if not predicate(obj):
            continue
        brs = obj.data_brs
        if brs is None:
            continue
        if obj.id is None:
            if not _nagged_ids:
                _logger.info("At least one FileObject had a null .id property.  Generating IDs.")
                _nagged_ids = True
            obj.id = _generate_id()
        else:
            if obj.id in _used_ids:
                _logger.warning("ID reuse: %r." % obj.id)
            _used_ids.add(obj.id)
        try:
            file_offset = 0
            cursor.execute("INSERT INTO files(obj_id, partition, inode, filename, filesize) VALUES (?,?,?,?,?);", (
              obj.id,
              obj.partition,
              obj.inode,
              obj.filename,
              obj.filesize
            ))
            found_incomplete_chunk = False
            for chunk in brs.iter_contents(raw_image, buffer_size=512):
                if found_incomplete_chunk:
                    _logger.debug("File with unexpected mid-stream incomplete byte run: %r." % obj)
                    raise ValueError("Found incomplete sector in middle of byte_runs list.")
                md5obj = hashlib.md5()
                sha1obj = hashlib.sha1()

                md5obj.update(chunk)
                sha1obj.update(chunk)

                if pad_sectors and len(chunk) < 512:
                    found_incomplete_chunk = True
                    remainder = 512 - len(chunk)
                    nulls = remainder * b"0"
                    md5obj.update(nulls)
                    sha1obj.update(nulls)

                #TODO No img_offset or fs_offset for now; could be done with a little byte_runs offset acrobatics, or a request to restore sector hash records in DFXML.
                cursor.execute("INSERT INTO block_hashes(obj_id, img_offset, fs_offset, file_offset, len, md5, sha1) VALUES (?,?,?,?,?,?,?);", (
                  obj.id,
                  None,
                  None,
                  file_offset,
                  len(chunk),
                  md5obj.hexdigest(),
                  sha1obj.hexdigest()
                ))

                file_offset += len(chunk)
            if not obj.filesize is None and file_offset != obj.filesize:
                _logger.warning("The hashed blocks' lengths do not sum to the filesize recorded: respectively, %d and %d.  File ID %r." % (file_offset, obj.filesize, obj.id))
        except AttributeError as e:
            #Some files' contents can't be accessed straightforwardly.  Note and skip.
            _logger.error(e.args[0] + ("  File ID %r." % obj.id))
            _logger.debug("The problem FileObject: %r." % obj)

        #Commit every thousand files
        if obj_no % 1000 == 999:
            _logger.debug("Committing hashes of object number %d." % obj_no)
            conn.commit()
    conn.commit()
    conn.close()

def is_allocated(fobj):
    if fobj.alloc_name and fobj.alloc_inode:
        return True
    elif fobj.alloc:
        return True
    return False

def is_new_file(fobj):
    if not fobj.annos is None:
        return  "new" in fobj.annos
    return None

def is_mod_file(fobj):
    if not fobj.annos is None:
        return  "modified" in fobj.annos
    return None

def is_new_or_mod_file(fobj):
    return is_new_file(fobj) or is_mod_file(fobj)

def main():
    predicates = {
      "all": (lambda x: True),
      "allocated": is_allocated,
      "new": is_new_file,
      "mod": is_mod_file,
      "newormod": is_new_or_mod_file
    }
    if args.predicate is None:
        args.predicate = "new"
    if args.predicate not in predicates:
        raise ValueError("--predicate must be from this list: %r.  Received: %r." % (predicates.keys(), args.predicate))

    if args.xml:
        d = Objects.parse(args.xml)
    else:
        d = Objects.parse(args.disk_image)
    write_sector_hashes_to_db(args.disk_image, d, is_allocated, args.db_output, args.pad)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description="Walks a file system and outputs sector hashes of all files matching a predicate.  Can be used as a library for the function write_sector_hashes_to_db.")
    parser.add_argument("-d", "--debug", action="store_true")
    parser.add_argument("-x", "--xml", help="Pre-computed DFXML.")
    parser.add_argument("--predicate", help="Condition for selecting files to sector hash.  One of 'new', 'allocated', 'all', 'mod'(ified), 'newormod'.  Default 'allocated'.")
    parser.add_argument("--pad", help="Pad non-full sectors with null bytes.", action="store_true")
    parser.add_argument("disk_image")
    parser.add_argument("db_output")
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    main()



#!/usr/bin/python
"""Reads an fiwalk XML file and reports how many of the files are still in the image..."""

import sys,os
sys.path.append(os.getenv("DOMEX_HOME") + "/src/lib/") # add the library
sys.path.append("../lib/")      # add the library
sys.path.append("../fiwalk/")

import fiwalk
import time

def calc_jumps(fis,title):
    print title
    print "Count: %d" % (len(fis))
    from histogram import histogram
    h = histogram()
    pos = 0
    backwards = 0
    prev_frag_count = 0
    for fi in fis:
        for i in range(0,len(fi.byte_runs())):
            run = fi.byte_runs()[i]
            try:
                sector = run.start_sector()
                if sector<pos:
                    backwards+=1
                    h.add((prev_frag_count,i))
                pos = sector
            except AttributeError:
                pass
        pref_frag_count = len(fi.byte_runs())
            
    print "Backwards Jumps: %d" % backwards
    print "Histogram of backwards:"
    h.print_top(10)

if __name__=="__main__":
    import sys
    from optparse import OptionParser
    from subprocess import Popen,PIPE
    global options

    parser = OptionParser()
    parser.add_option("-d","--debug",help="prints debugging info",dest="debug")
    parser.add_option("-x","--xmlfile",help="XML file (optional)")
    parser.add_option("-i","--imagefile",help="image file (required)")
    parser.usage = '%prog [options] xmlfile diskimage'
    (options,args) = parser.parse_args()

    if not options.xmlfile or not options.imagefile:
        parser.print_help()
        sys.exit(1)

    # Read the redaction configuration file
    imagefile = open(options.imagefile,"r")
    if options.xmlfile:
        xmlfile   = open(options.xmlfile,"r")
    else:
        xmlfile   = None

    t0  = time.time()
    fis = fiwalk.fileobjects_using_sax(imagefile=imagefile,xmlfile=xmlfile)
    t1  = time.time()
    print("Time to read file objects: {} seconds".format(t1-t0))

    # Create a new array with just those that we can read
    def resident_file(fi):
        if len(fi.byte_runs())==0: return False
        if len(fi.byte_runs())>2: return False
        if hasattr(fi.byte_runs()[0],"uncompressed_len"): return False
        if not hasattr(fi.byte_runs()[0],"img_offset"): return False
        return True

    fis = filter(resident_file,fis)

    print "Native order: "
    calc_jumps(fis,"Native Order")
    def sort_function(a,b):
        a0 = a.byte_runs()[0].start_sector()
        b0 = b.byte_runs()[0].start_sector()
        if a0 < b0 : return -1
        if a0 == b0 : return 0
        return 1
    fis.sort(sort_function)
    calc_jumps(fis,"Sorted Order")
    





#!/usr/bin/env python3

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain. NIST assumes no
# responsibility whatsoever for its use by other parties, and makes
# no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""Walk current directory, writing DFXML to stdout."""

__version__ = "0.2.1"

import os
import stat
import hashlib
import traceback
import logging
import sys

_logger = logging.getLogger(os.path.basename(__file__))

import Objects

def filepath_to_fileobject(filepath):
    fobj = Objects.FileObject()

    #Determine type - done in three steps.
    if os.path.islink(filepath):
        fobj.name_type = "l"
    elif os.path.isdir(filepath):
        fobj.name_type = "d"
    elif os.path.isfile(filepath):
        fobj.name_type = "r"
    else:
        #Need to finish type determinations with stat structure.
        pass

    #Prime fileobjects from Stat data (lstat for soft links).
    if fobj.name_type == "l":
        sobj = os.lstat(filepath)
    else:
        sobj = os.stat(filepath)
    #_logger.debug(sobj)
    fobj.populate_from_stat(sobj)

    if fobj.name_type is None:
        if stat.S_ISCHR(fobj.mode):
            fobj.name_type = "c"
        elif stat.S_ISBLK(fobj.mode):
            fobj.name_type = "b"
        elif stat.S_ISFIFO(fobj.mode):
            fobj.name_type = "p"
        elif stat.S_ISSOCK(fobj.mode):
            fobj.name_type = "s"
        elif stat.S_ISWHT(fobj.mode):
            fobj.name_type = "w"
        else:
            raise NotImplementedError("No reporting check written for file type of %r." % filepath)

    #Hard-coded information: Name, and assumed allocation status.
    fobj.filename = filepath
    fobj.alloc = True

    if fobj.name_type == "l":
        fobj.link_target = os.readlink(filepath)

    #Add hashes for regular files.
    if fobj.name_type == "r":
        try:
            with open(filepath, "rb") as in_fh:
                chunk_size = 2**22
                md5obj = hashlib.md5()
                sha1obj = hashlib.sha1()
                any_error = False
                while True:
                    buf = b""
                    try:
                        buf = in_fh.read(chunk_size)
                    except Exception as e:
                        any_error = True
                        fobj.error = "".join(traceback.format_stack())
                        if e.args:
                            fobj.error += "\n" + str(e.args)
                        buf = b""
                    if buf == b"":
                        break

                    md5obj.update(buf)
                    sha1obj.update(buf)

                if not any_error:
                    fobj.md5 = md5obj.hexdigest()
                    fobj.sha1 = sha1obj.hexdigest()
        except Exception as e:
            if fobj.error is None:
                fobj.error = ""
            else:
                fobj.error += "\n"
            fobj.error += "".join(traceback.format_stack())
            if e.args:
                fobj.error += "\n" + str(e.args)
    return fobj

def main():
    #Determine whether we're going in threading mode or not.  (Some modules are not available by default.)
    using_threading = False
    if args.jobs > 1:
        using_threading = True  #(unless supporting modules are absent)
        try:
            import threading
        except:
            using_threading = False
            _logger.warning("Threading support not available.  Running in single thread only.")

        try:
            import queue
        except:
            using_threading = False
            _logger.warning("Python queue support not available.  (If running Ubuntu, this is in package python3-queuelib.)  Running in single thread only.")

    dobj = Objects.DFXMLObject(version="1.1.1")
    dobj.program = sys.argv[0]
    dobj.program_version = __version__
    dobj.command_line = " ".join(sys.argv)
    dobj.dc["type"] = "File system walk"

    filepaths = set()
    filepaths.add(".")
    for (dirpath, dirnames, filenames) in os.walk("."):
        dirent_names = set()
        for dirname in dirnames:
            dirent_names.add(dirname)
        for filename in filenames:
            dirent_names.add(filename)
        for dirent_name in sorted(dirent_names):
            #The relpath wrapper removes "./" from the head of the path.
            filepath = os.path.relpath(os.path.join(dirpath, dirent_name))
            filepaths.add(filepath)

    fileobjects_by_filepath = dict()

    if using_threading:
        #Threading syntax c/o: https://docs.python.org/3.5/library/queue.html
        q = queue.Queue()
        threads = []

        def _worker():
            while True:
                filepath = q.get()
                if filepath is None:
                    break
                fobj = filepath_to_fileobject(filepath)
                fileobjects_by_filepath[filepath] = fobj
                q.task_done()

        for i in range(args.jobs):
            t = threading.Thread(target=_worker)
            t.start()
            threads.append(t)

        for filepath in filepaths:
            q.put(filepath)

        # block until all tasks are done
        q.join()

        # stop workers
        for i in range(args.jobs):
            q.put(None)
        for t in threads:
            t.join()
    else: #Not threading.
        for filepath in sorted(filepaths):
            fobj = filepath_to_fileobject(filepath)
            fileobjects_by_filepath[filepath] = fobj

    #Build output DFXML tree.
    for filepath in sorted(fileobjects_by_filepath.keys()):
        dobj.append(fileobjects_by_filepath[filepath])
    dobj.print_dfxml()

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", action="store_true")
    parser.add_argument("-j", "--jobs", type=int, default=1, help="Number of file-processing threads to run.")
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    if args.jobs <= 0:
        raise ValueError("If requesting multiple jobs, please request 1 or more worker threads.")

    main()



#
# Report the difference between two dfxml files
#
from filesdb import filesdb
import dfxml
import sys
#
# test program. Reads a database and dumps it.
#
if __name__=="__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(description='Test the files database with one or more DFXML files')
    parser.add_argument('xmlfiles',help='XML files to process',nargs='+')

    args = parser.parse_args()
    db0   = None
    for fn in args.xmlfiles:
        db1 = filesdb()
        db1.fname = fn
        db1.read(fn)
        print("{} stats:".format(fn))
        db1.print_stats(sys.stdout)
        if db0:
            print("")
            print("Difference from {}".format(db0.fname))
        db0 = db1





#!/usr/bin/env python
### DO NOT MODIFY THIS FILE ###
### DOWNLOAD NEW FILE FROM https://raw.github.com/simsong/dfxml/master/python/fiwalk.py
#
# fiwalk version 0.6.3
#
# %%% BEGIN NO FILL
"""fiwalk module

This is the part of dfxml that is dependent on fiwalk.py

"""
import dfxml
from sys import stderr
from subprocess import Popen,PIPE
ALLOC_ONLY = 1

fiwalk_cached_installed_version = None
def fiwalk_installed_version(fiwalk='fiwalk'):
    """Return the current version of fiwalk that is installed"""
    global fiwalk_cached_installed_version
    if fiwalk_cached_installed_version:
        return fiwalk_cached_installed_version
    from subprocess import Popen,PIPE
    import re
    for line in Popen([fiwalk,'-V'],stdout=PIPE).stdout.read().decode('utf-8').split("\n"):
        g = re.search("^FIWalk Version:\s+(.*)$",line)
        if g:
            fiwalk_cached_installed_version = g.group(1)
            return fiwalk_cached_installed_version
        g = re.search("^SleuthKit Version:\s+(.*)$",line)
        if g:
            fiwalk_cached_installed_version = g.group(1)
            return fiwalk_cached_installed_version
    return None

class XMLDone(Exception):
    def __init__(self,value):
        self.value = value

class version:
    def __init__(self):
        self.cdata = ""
        self.in_element = []
        self.version = None
    def start_element(self,name,attrs):
        if(name=='volume'):     # too far?
            raise XMLDone(None)
        self.in_element += [name]
        self.cdata = ""
    def end_element(self,name):
        if ("fiwalk" in self.in_element) and ("creator" in self.in_element) and ("version" in self.in_element):
            raise XMLDone(self.cdata)
        if ("fiwalk" in self.in_element) and ("fiwalk_version" in self.in_element):
            raise XMLDone(self.cdata)
        if ("version" in self.in_element) and ("dfxml" in self.in_element) and ("creator" in self.in_element):
            raise XMLDone(self.cdata)
        self.in_element.pop()
        self.cdata = ""
    def char_data(self,data):
        self.cdata += data

    def get_version(self,fn):
        import xml.parsers.expat
        p = xml.parsers.expat.ParserCreate()
        p.StartElementHandler  = self.start_element
        p.EndElementHandler    = self.end_element
        p.CharacterDataHandler = self.char_data
        try:
            p.ParseFile(open(fn,'rb'))
        except XMLDone as e:
            return e.value
        except xml.parsers.expat.ExpatError:
            return None             # XML error
        
def fiwalk_xml_version(filename=None):
    """Returns the fiwalk version that was used to create an XML file.
    Uses the "quick and dirty" approach to getting to getting out the XML version."""

    p = version()
    return p.get_version(filename)

################################################################
def E01_glob(fn):
    import os.path
    """If the filename ends .E01, then glob it. Currently only handles E01 through EZZ"""
    ret = [fn]
    if fn.endswith(".E01") and os.path.exists(fn):
        fmt = fn.replace(".E01",".E%02d")
        for i in range(2,100):
            f2 = fmt % i
            if os.path.exists(f2):
                ret.append(f2)
            else:
                return ret
        # Got through E99, now do EAA through EZZ
        fmt = fn.replace(".E01",".E%c%c")
        for i in range(0,26):
            for j in range(0,26):
                f2 = fmt % (i+ord('A'),j+ord('A'))
                if os.path.exists(f2):
                    ret.append(f2)
                else:
                    return ret
        return ret              # don't do F01 through F99, etc.
    return ret


def fiwalk_xml_stream(imagefile=None,flags=0,fiwalk="fiwalk",fiwalk_args=""):
    """ Returns an fiwalk XML stream given a disk image by running fiwalk."""
    if flags & ALLOC_ONLY: fiwalk_args += "-O"
    from subprocess import call,Popen,PIPE
    # Make sure we have a valid fiwalk
    try:
        res = Popen([fiwalk,'-V'],stdout=PIPE).communicate()[0]
    except OSError:
        raise RuntimeError("Cannot execute fiwalk executable: "+fiwalk)
    cmd = [fiwalk,'-x']
    if fiwalk_args: cmd += fiwalk_args.split()
    p = Popen(cmd + E01_glob(imagefile.name),stdout=PIPE)
    return p.stdout

def fiwalk_using_sax(imagefile=None,xmlfile=None,fiwalk="fiwalk",flags=0,callback=None,fiwalk_args=""):
    """Processes an image using expat, calling a callback for every file object encountered.
    If xmlfile is provided, use that as the xmlfile, otherwise runs fiwalk."""
    import dfxml
    if xmlfile==None:
        xmlfile = fiwalk_xml_stream(imagefile=imagefile,flags=flags,fiwalk=fiwalk,fiwalk_args=fiwalk_args)
    r = dfxml.fileobject_reader(flags=flags)
    r.imagefile = imagefile
    r.process_xml_stream(xmlfile,callback)

def fileobjects_using_sax(imagefile=None,xmlfile=None,fiwalk="fiwalk",flags=0):
    ret = []
    fiwalk_using_sax(imagefile=imagefile,xmlfile=xmlfile,fiwalk=fiwalk,flags=flags,
                     callback = lambda fi:ret.append(fi))
    return ret

def fileobjects_using_dom(imagefile=None,xmlfile=None,fiwalk="fiwalk",flags=0,callback=None):
    """Processes an image using expat, calling a callback for every file object encountered.
    If xmlfile is provided, use that as the xmlfile, otherwise runs fiwalk."""
    import dfxml
    if xmlfile==None:
        xmlfile = fiwalk_xml_stream(imagefile=imagefile,flags=flags,fiwalk=fiwalk)
    return dfxml.fileobjects_dom(xmlfile=xmlfile,imagefile=imagefile,flags=flags)

ctr = 0
def cb_count(fn):
    global ctr
    ctr += 1

if __name__=="__main__":
    import sys
    for fn in sys.argv[1:]:
        print("{} contains fiwalk version {}".format(fn,fiwalk_xml_version(fn)))
        # Count the number of files
        fiwalk_using_sax(xmlfile=open(fn,'rb'),callback=cb_count)
        print("Files: {}".format(ctr))



#!/usr/bin/env python

# This software was developed at the National Institute of Standards
# and Technology in whole or in part by employees of the Federal
# Government in the course of their official duties. Pursuant to
# title 17 Section 105 of the United States Code portions of this
# software authored by NIST employees are not subject to copyright
# protection and are in the public domain. For portions not authored
# by NIST employees, NIST has been granted unlimited rights. NIST
# assumes no responsibility whatsoever for its use by other parties,
# and makes no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

__version__ = "0.1.0"

import Objects

def main():
    dobj = Objects.parse(args.in_dfxml)
    assert dobj.program == args.expected_program
    assert dobj.program_version == args.expected_program_version

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("in_dfxml")
    parser.add_argument("expected_program")
    parser.add_argument("expected_program_version")
    args = parser.parse_args()

    main()



#
# Using two XML files make the current system look like the master
#
from filesdb import filesdb
import dfxml
import sys
#
# test program. Reads a database and dumps it.
#
if __name__=="__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(description='Make the local system look like the master')
    parser.add_argument('--commit',help='Actually do the job',action='store_true')
    parser.add_argument('--local',help='speciies an XML file that describes the local system (required)')
    parser.add_argument('masterfiles',
                        help='XML files to process. Files may be prefixed with an [xml] path',
                        nargs='+')

    args = parser.parse_args()

    if not args.local:
        parser.print_help()
        exit(1)

    masterdb   = filesdb()
    for fn in args.masterfiles:
        masterdb.read_with_prefix(fn)
    print("Master stats:")
    masterdb.print_stats(sys.stdout)
    print("\n")
    print("Local mirror stats:")
    ldb = filesdb()
    ldb.read_with_prefix(args.local)

    # Create new directories if needed
    for newdir in ldb.new_dirs(masterdb):
        print("mkdir {}".format(newdir))
    
    keep_files = []
    mv_files   = []
    rm_files   = []
    def process_fi(fi):
        # If hash is same and name is the same, ignore:
        for nfi in masterdb.search(fi,hash=True,name=True):
            keep_files.append(fi.filename())
            return              # in the database
        
        # If hash is same and name is different, move it
        for nfi in masterdb.search(fi,hash=True):
            mv_files.append((fi.filename(),nfi.filename()))
            return
        
        # If name is same and hash is different, erase it
        for nfi in masterdb.search(fi,name=True):
            rm_files.append(fi.filename())
            return

        # Otherwise, erase the hash
        rm_files.append(fi.filename())
        return

    # Rename files that need to be renamed
    for fi in ldb:
        process_fi(fi)

    print("Files to keep: {:12,}".format(len(keep_files)))
    print("Files to rm:   {:12,}".format(len(rm_files)))
    print("Files to mv:   {:12,}".format(len(mv_files)))
    



#
# filesdb
# a module that holds a database of DFXML files
#

import dfxml
from collections import defaultdict
import sys
class filesdb:
    def __init__(self,fname=None):
        self.sha1db   = defaultdict(list) # fi's by hashdb
        self.md5db    = defaultdict(list) # fi's by hashdb
        self.fnamedb  = defaultdict(list) # fi's by fname
        self.dirs     = defaultdict(list) # fi's by directory name
        self.fis      = []  
        self.prefix   = None
        self.delfix   = None
        if fname:
            self.read(fname)
    
    def __iter__(self):
        """The iterator for filesdb iterates through all the files"""
        return self.fis.__iter__()

    def read(self,f):
        if type(f)==str:
            self.fname = f
            f = open(f,'rb')
        dfxml.read_dfxml(xmlfile=f,callback=self.pass1)

    def read_with_prefix(self,fname):
        if ':' in fname:
            (fmt,fname) = fname.split(':')
            if fmt[0]=='+': self.prefix = fmt[1:]
            if fmt[0]=='=': self.delfix = fmt[1:]
            if fmt[0]!='+' and fmt[0]!='=': self.prefix = fmt
        self.read(fname)

    def pass1(self,fi):
        """First pass for reading fi objects"""
        import os
        self.fis.append(fi)
        if fi.sha1(): self.sha1db[fi.sha1()].append(fi)
        if fi.md5(): self.md5db[fi.md5()].append(fi)
        if fi.filename():
            fname = fi.filename()
            if self.delfix:
                if fname.startswith(self.delfix): fname = fname[len(self.delfix):]
            if self.prefix:
                fname = self.prefix + fname
            self.sha1db[fname].append(fi)
            self.dirs[os.path.dirname(fname)].append(fi)
        
    def print_stats(self,f=sys.stdout):
        """Returns a text string of the stats"""
        ret = [
            ['Total directories',len(self.dirs)],
            ['Total files',len(self.fis)],
            ['Total bytes',sum([int(fi.filesize()) for fi in self.fis])],
            ['Total sha1s',len(self.sha1db)],
            ['Total md5s',len(self.md5db)],
            ]
        print("\n".join(["{:20}: {:14,}".format(a[0],a[1]) for a in ret]))

        mtime_min = [fi.mtime() for fi in self.fis]
        #print('mtime=',len(mtime_min))
        #flt = list(filter(lambda a:a!=None,mtime_min))
        #print('flt=',flt,len(flt))

        #print('mtime_min=',mtime_min)
        #print(['ctime range',mtime_min])
        #exit(0)


    def del_dirs(self,targetdb):
        """Given a targetdb, provide the dirs to get there."""
        return set(self.dirs.keys()).difference(set(targetdb.dirs.keys()))
        
    def del_files(self,targetdb):
        """Given an targetdb, provide the files needed to get there."""
        return set(self.filesdb).difference(set(db.filesdb))
        
    def new_dirs(self,db):
        """Given an older db, provide the dirs that are new."""
        return set(db.dirs.keys()).difference(set(self.dirs.keys()))
        
    def search(self,mfi,hash=False,name=False):
        """Return the matching fis"""
        if hash and not name:
            return self.md5db[mfi.md5()]
        if name and not hash:
            return self.fnamedb[mfi.filename()]
        if hash and name:
            return filter(lambda fi:fi.filename()==mfi.filename(),self.md5db[mfi.md5()])
        return []

#
# test program. Reads a database and dumps it.
#
if __name__=="__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(description='Test the files database with one or more DFXML files')
    parser.add_argument('xmlfiles',help='XML files to process',nargs='+')

    args = parser.parse_args()
    db   = filesdb()
    for fn in args.xmlfiles:
        db.read(fn)
    print(db.stats())




import Objects
import copy
import logging
import os

_logger = logging.getLogger(os.path.basename(__file__))
logging.basicConfig(level=logging.DEBUG)

ho = Objects.HiveObject()

ho.mtime = "2010-01-02T03:45:00Z"

hoc = copy.deepcopy(ho)

diffs = hoc.compare_to_other(ho)
_logger.debug(repr(diffs))
assert len(diffs) == 0

hoc.mtime = "2011-01-02T03:45:00Z"

diffs = hoc.compare_to_other(ho)
_logger.debug(repr(diffs))
assert len(diffs) == 1



#!/usr/bin/env python
"""tcpdiff.py

Generates a report about what's different between two tcp DFXML files
produced by tcpflow.

Process:

"""

import sys,time
if sys.version_info < (3,1):
    raise RuntimeError("rdifference.py requires Python 3.1 or above")

import fiwalk,dfxml,dfxml_html
def ptime(t):
    """Print the time in the requested format. T is a dfxml time value"""
    global options
    if t is None:
        return None
    elif options.timestamp:
        return str(t.timestamp())
    else:
        return str(t.iso8601())

def dprint(x):
    "Debug print"
    global options
    if options.debug: print(x)

#
# This program keeps track of the current and previous TCP connections in a single
# object called "FlowState". Another way to do that would have been to have
# the instance built from the XML file and then have another function that compares
# them.
#        

class FlowState:
    def __init__(self,fname):
        self.options = options
        self.connections = set()
        self.process(fname)
        
    def process(self,fname):
        self.fname = fname
        dfxml.read_dfxml(xmlfile=open(fname,'rb'), callback=self.process_fi)

    def process_fi(self,fi):
        self.connections.add(fi)

    def report(self):
        dfxml_html.header()
        dfxml_html.h1("DFXML file:"+self.current_fname)
        dfxml_html.table(['Total Connections',str(len(self.connections))])

if __name__=="__main__":
    from optparse import OptionParser
    from copy import deepcopy
    global options

    parser = OptionParser()
    parser.usage = '%prog [options] file1 file2 (files MUST be tcpflow DFXML files)'
    parser.add_option("-d","--debug",help="debug",action='store_true')

    (options,args) = parser.parse_args()

    if len(args)!=2:
        parser.print_help()
        sys.exit(1)

    a = FlowState(fname=args[0])
    a.report()

    b = FlowState(fname=args[1])
    b.report()

    print("Difference:")



#!/usr/bin/env python
# coding=UTF-8
"""dfxinfo.py:
Generates a report about what up with a DFXML file.
"""

import platform,os,os.path,sys,time
if platform.python_version_tuple() < ('3','2','0'):
    raise RuntimeError('This script now requires Python 3.2 or above')

try:
    import dfxml, fiwalk
except ImportError:
    raise ImportError('This script requires the dfxml and fiwalk modules for Python.')

__version__='0.0.1'

import fiwalk,dfxml
from histogram import histogram

class DiskSet:
    """DiskSet maintains a database of the file objects within a disk.
    The entire database is maintained in memory."""
    def __init__(self):
        self.ext_histogram          = histogram()
        self.ext_histogram_distinct = histogram()
        self.fi_by_md5              = dict() # a dictionary of lists

    def pass1(self,fi):
        if fi.is_virtual(): return
        if fi.is_file():
            self.fi_by_md5.setdefault(fi.md5(),[]).append(fi)

    def print_dups_report(self):
        print("Duplicates:")
        # First extract the dups, then sort them
        dups  = filter(lambda a:len(a[1])>1,self.fi_by_md5.items(),)
        dup_bytes = 0
        for (md5hash,fis) in sorted(dups,key=lambda a:a[1][0].filesize(),reverse=True):
            for fi in fis:
                print("{:>16,} {:32} {}".format(fi.filesize(),fi.md5(),fi.filename()))
            print()
            dup_bytes += fis[0].filesize() * (len(fis)-1)
        print("Total duplicate bytes: {:,}".format(dup_bytes))


if __name__=="__main__":
    from argparse import ArgumentParser
    from copy import deepcopy

    parser = ArgumentParser(description='Report information about a DFXML file')
    parser.add_argument('xmlfiles',help='XML files to process',nargs='+')
    parser.add_argument("--imagefile",help="specifies imagefile to examine; automatically runs fiwalk",nargs='+')

    args = parser.parse_args()
    ds   = DiskSet()
    for fn in args.xmlfiles:
        print("Processing {}".format(fn))
        dfxml.read_dfxml(xmlfile=open(fn,'rb'),callback=ds.pass1)
    ds.print_dups_report()



#!/usr/bin/python
#
# deidentify_xml.py:
# Given XML for a disk, remove information that might be personally identifying from filenames.
# remember the mapping so that directory names don't get changed.
#
# 2012-10-27 slg - updated to Python3

private_dirs = ["home/","usr/home","Users"]
ok_top_paths_win = ["program files/","System","Windows"]
ok_top_paths_mac = ["bin/","usr","etc","private","applications","developer",'bin','sbin','lib','dev']
ok_top_paths = ok_top_paths_win + ok_top_paths_mac + ['$orphanfiles']
acceptable_extensions = ["exe","dll","sys","com","hlp"]

import os.path, os, sys

partdir = {}
def sanitize_part(part):
    """Sanitize a part of a pathname in a consistent manner"""
    if part not in partdir:
        partdir[part] = "P%07d" % (len(partdir)+1)
    return partdir[part]

def sanitize_filename(fname):
    """Given a filename, sanitize each part and return it."""
    ofn = fname
    jfn = fname
    if jfn[0]=='/': jfn=jfn[1:]
    pathok = False
    for p in ok_top_paths:
        if jfn.lower().startswith(p):
            pathok = True

    if not pathok:
        # if the path is not okay, replace all of the parts
        # and the name up to the .ext
        parts = fname.split("/")
        parts[:-1] = [sanitize_part(s) for s in parts[:-1]]
        (root,ext) = os.path.splitext(parts[-1])
        if ext not in acceptable_extensions:
            parts[-1] = sanitize_part(root) + ext
        fname = "/".join(parts)
    if ofn[0]=='/' and fname[0]!='/':
        fname = "/" + fname
    return fname
    
class xml_sanitizer:
    """Read and write the XML, but sanitize the filename elementss."""
    def __init__(self,out):
        self.out = out
        self.cdata = ""
        
    def _start_element(self, name, attrs):
        """ Handles the start of an element for the XPAT scanner"""
        s = ['<',name]
        if attrs:
            for (a,v) in attrs.items():
                if '"' not in v:
                    s += [' ',a,'="',v,'"']
                else:
                    s += [" ",a,"='",v,"'"]
        s += ['>']
        self.out.write("".join(s))
        self.cdata = ""                 # new element

    def _end_element(self, name):
        """Handles the end of an element for the XPAT scanner"""
        if name=="filename":
            self.cdata = sanitize_filename(self.cdata)
        if self.cdata=="\n": self.cdata=""
        self.out.write("".join([self.cdata,'</',name,'>']))
        self.cdata = ""

    def _char_data(self, data):
        """Handles XML data"""
        self.cdata += data

    def process_xml_stream(self,xml_stream):
        "Run the reader on a given XML input stream"
        import xml.parsers.expat
        p = xml.parsers.expat.ParserCreate()
        p.StartElementHandler  = self._start_element
        p.EndElementHandler    = self._end_element
        p.CharacterDataHandler = self._char_data
        p.ParseFile(xml_stream)

if __name__=="__main__":
    from optparse import OptionParser
    global options
    parser = OptionParser()
    parser.add_option("-t","--test",help='Test a specific pathanme to sanitize')
    (options,args) = parser.parse_args()

    if options.test:
        if os.path.isdir(options.test):
            for (dirpath,dirnames,filenames) in os.walk(options.test):
                for filename in filenames:
                    fn = dirpath+"/"+filename
                    print("%s\n   %s" % (fn,sanitize_filename(fn)))

    x = xml_sanitizer(sys.stdout)
    x.process_xml_stream(open(args[0],'rb'))
    
            
            



#!/usr/bin/env python3

# This software was developed at the National Institute of Standards
# and Technology in whole or in part by employees of the Federal
# Government in the course of their official duties. Pursuant to
# title 17 Section 105 of the United States Code portions of this
# software authored by NIST employees are not subject to copyright
# protection and are in the public domain. For portions not authored
# by NIST employees, NIST has been granted unlimited rights. NIST
# assumes no responsibility whatsoever for its use by other parties,
# and makes no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

"""
Writes to stdout a DFXML document that is the concatenation of each input DFXML document.

Assumes the input DFXML has at most one volume per document.

Partition numbers, offsets, and byte run img_offset attributes are all overwritten with the correct value assuming the partition was carved at the offset passed on the command line.

That is, this command:

    $0 32256:fiout.xml 1073741824:fiout.xml

will create a single DFXML file with two volumes and their file objects contained.
"""

__version__ = "0.2.0"

import Objects
import logging
import os
import sys
import xml.etree.ElementTree as ET

_logger = logging.getLogger(os.path.basename(__file__))

def main():
    d = Objects.DFXMLObject(version="1.1.1")

    d.program = sys.argv[0]
    d.program_version = __version__
    d.command_line = " ".join(sys.argv)

    _offsets_and_pxml_paths = []
    for (lxfno, lxf) in enumerate(args.labeled_xml_file):
        lxf_parts = lxf.split(":")
        if len(lxf_parts) != 2 or not lxf_parts[0].isdigit():
            raise ValueError("Malformed argument in labeled_xml_file.  Expecting space-delimited list of '<number>:<path>'.  This entry doesn't work: %r." % lxf)
        offset = int(lxf_parts[0])
        path = lxf_parts[1]
        _offsets_and_pxml_paths.append((offset,path))
    offsets_and_pxml_paths = sorted(_offsets_and_pxml_paths)

    for (pxml_path_index, (offset, pxml_path)) in enumerate(offsets_and_pxml_paths):
        _logger.debug("Running on path %r." % pxml_path)
        pdo = Objects.parse(pxml_path)

        building_volume = None
        #Fetch or build volume we'll append
        if len(pdo.volumes) > 1:
            raise ValueError("An input DFXML document has multiple volumes; this script assumes each input document only has one.  The document here has %d: %r." % (len(pdo.volumes), pxml_path))
        elif len(pdo.volumes) == 0:
            v = Objects.VolumeObject()
            building_volume = True
        else:
            v = pdo.volumes[0]
            building_volume = False

        v.partition_offset = offset

        #Accumulate namespaces
        for (prefix, url) in pdo.iter_namespaces():
            d.add_namespace(prefix, url)

        for obj in pdo:
            #Force-update image offsets in byte runs
            for brs_prop in ["data_brs", "name_brs", "inode_brs"]:
                if hasattr(obj, brs_prop):
                    brs = getattr(obj, brs_prop)
                    if brs is None:
                        continue
                    for br in brs:
                        if not br.fs_offset is None:
                            br.img_offset = br.fs_offset + offset
            #For files, set partition identifier and attach to partition
            if isinstance(obj, Objects.FileObject):
                obj.partition = pxml_path_index + 1
                if building_volume:
                    v.append(obj)

        #Collect the constructed and/or updated volume
        d.append(v)

    d.print_dfxml()

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", help="Enable debug printing", action="store_true")
    parser.add_argument("--image-path", help="Path to the source image file to record in the resulting DFXML.")
    parser.add_argument("labeled_xml_file", help="List of DFXML files, each colon-prefixed with the partition's offset in bytes (e.g. '32256:fiout.dfxml')", nargs="+")
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    main()



#!/usr/bin/python
"""Usage: iblkfind imagefile.iso s1 [s2 s3 ...] ...

Reports the files in which sectors s1, s2, s3... are located.
"""
import dfxml,sys


if __name__=="__main__":
    from optparse import OptionParser
    parser = OptionParser()
    parser.usage = '%prog [options] imagefile-or-xmlfile s1 [s2 s3 s3 ...]'
    parser.add_option("--offset",help="values are byte offsets, not sectors",action="store_true")
    parser.add_option("--blocksize",help="specify sector blockszie",default=512)
    (options,args) = parser.parse_args()

    if len(args)<1:
        parser.print_help()
        sys.exit(1)
    fn = args[0]

    print(args)
    print("Processing %s" % fn)
    print("Searching for %s" % ", ".join(args[1:]))

    divisor = 1
    if options.offset:
        divisor = options.blocksize

    sectors = set([int(s)/divisor for s in args[1:]])


    def process(fi):
        for s in sectors:
            if fi.has_sector(s):
                print("%d\t%s" % (s,fi.filename()))
    
    if not fn.endswith(".xml"):
        print("iblkfind requires an XML file")
        exit(1)
    dfxml.read_dfxml(xmlfile=open(args[0],"rb"),callback=process)




import Objects
import test_diffing_CellObject
import test_diffing_HiveObject

ro = Objects.RegXMLObject(version="0.2")
ho = Objects.HiveObject()
ho.append(test_diffing_CellObject.co)
ho.append(test_diffing_CellObject.nco)
ro.append(test_diffing_HiveObject.ho)
ro.print_regxml()



#!/usr/bin/python
"""Usage: imap imagefile0.iso imagefile1.iso imagefile2.iso ...

Produces a map of imagefile0.iso, using the other image files as "hints" for missing
data. Only reports files that have been allocated; deleted files are reported only if
they can be found allocated in another file.
"""


import fiwalk


################################################################
if __name__=="__main__":
    import sys
    from sys import stdout

    from optparse import OptionParser
    parser = OptionParser()
    parser.usage = '%prog [options] image.iso '
    parser.add_option("-d","--debug",help="debug",action="store_true")
    (options,args) = parser.parse_args()

    if len(args)<1:
        parser.print_help()
        sys.exit(1)

    imagefile = open(args[0],"r")
    annotated_runs = []
    if options.debug: print("Read %d file objects from %s" % (len(fileobjects),imagefile.name))

    def cb(fi):
        if options.debug: print("Read "+str(fi))
        fragment_num = 1
        for run in fi.byte_runs():
            annotated_runs.append((run.img_offset,run,fragment_num,fi))
            fragment_num += 1
    fiwalk.fiwalk_using_sax(imagefile=imagefile,callback=cb)

    next_sector = 0

    for (ip,run,fragment_num,fi) in sorted(annotated_runs):
        extra = ""
        fragment = ""
        start_sector = run.img_offset/512
        sector_count = int(run.bytes/512)
        partial        = run.bytes % 512
    
        if not fi.allocated():
            print("***")

        if not fi.file_present():       # it's not here!
            continue 

        if partial>0:
            sector_count += 1
            extra = "(%3d bytes slack)" % (512-partial)

        if fi.fragments()>2:
            fragment = "fragment %d" % fragment_num

        if next_sector != start_sector:
            print "  <-- %5d unallocated sectors @ sector %5d -->" % (start_sector-next_sector,next_sector)

        print("[ %6d  -> %6d sectors %18s ]   %s  %s " % (start_sector,sector_count,extra,fi.filename(),fragment))

        next_sector = start_sector + sector_count




#!/usr/bin/env python
import dfxml, sys

timeline = []

def process(co):
    mtime = co.mtime()
    if mtime != None:
        timeline.append([co.mtime(),co.full_path()," modified"])

def main():
    if len(sys.argv) < 2:
        print("Usage: {} <input.regxml>".format(sys.argv[0]))
        exit(1)
    dfxml.read_regxml(xmlfile=open(sys.argv[1],"rb"), callback=process)
    timeline.sort()
    for record in timeline:
        print("\t".join( map(str, record)) )

if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Print the stats from a DFXML file 

import sys,os,shelve

sys.path.append(os.getenv("HOME")+"/slg/src/python")
sys.path.append(os.getenv("DOMEX_HOME")+"/src/lib")

from histogram2d import histogram2d
from histogram import histogram
from statbag import statbag
from ttable import ttable
import re,dfxml,fiwalk

def process_files(fn):
    drive_files = {}                         # index of drives
    all_parts  = []
    all_files = []
    files_by_md5 = {}           # a dictionary of sets of fiobject, indexed by md5
    extension_len_histogram = histogram2d()
    extension_fragments_histogram = histogram2d()
    partition_histogram = histogram2d()

    def cb(fi):
        # add the md5 to the set
        if fi.is_file() and fi.filesize():
            files_by_md5.get(fi.md5,set()).add(fi)
            ext = fi.ext()
            if not ext: print fi.meta_type(),fi
            extension_len_histogram.add(ext,fi.filesize())
            extension_fragments_histogram.add(ext,fi.fragments())
            partition_histogram.add(fi.partition(),fi.filesize())

    if fn.endswith('xml'):
        fiwalk.fiwalk_using_sax(xmlfile=open(fn),callback=cb)
    else:
        fiwalk.fiwalk_using_sax(imagefile=open(fn),callback=cb)
    

    #
    # Typeset the information
    #

    tab = ttable()
    tab.header     = "File extension popularity and average size (suppressing 0-len files)"
    tab.col_headings = [['Ext','Count','Average Size','Max','Std Dev']]
    tab.omit_row = [[0,'']]
    extension_len_histogram.statcol = ['iaverage','maxx','istddev']
    print extension_len_histogram.typeset(tab=tab)

    #
    # Information about fragmentation patterns
    #
    tab = ttable()
    tab.header="Fragmentation pattern by file system and file type:"
    tab.col_headings = [['Ext','Count','Average Size','Max','Std Dev']]
    tab.omit_row = [[0,'']]
    extension_fragments_histogram.statcol = ['iaverage','maxx','istddev']
    print extension_fragments_histogram.typeset(tab=tab)
    exit(0)

    for fstype in fstypes:
        for ftype in ['jpg','pdf','doc','txt']:
            len1stats = statbag()
            len2stats = statbag()
            delta_hist = histogram()
            delta_re = re.compile("(\d+)\-?(\d+)? ?(\d+)\-?(\d+)?")
            for i in filter( (lambda(f): f.ext()==ftype and f.fragments==2),all_files):
                runs = False
                if(hasattr(i,'block_runs')): runs = i.block_runs
                if(hasattr(i,'sector_runs')): runs = i.sector_runs
                if not runs: continue
                m = delta_re.search(runs)
                r = []
                for j in range(1,5):
                    try:
                        r.append(int(m.group(j)))
                    except TypeError:
                        r.append(int(m.group(j-1)))

                len1 = r[1] - r[0] + 1
                len2 = r[3] - r[2] + 1
                delta = r[2]-r[1]
                
                len1stats.addx(len1)
                len2stats.addx(len2)
                delta_hist.add(delta)

            if len1stats.count()>0:
                print "\n\n"
                print "fstype:",fstype,"  ftype:",ftype
                print "len1 average: %f stddev: %f" % (len1stats.average(),len1stats.stddev())
                print "len2 average: %f stddev: %f" % (len2stats.average(),len2stats.stddev())
                print "delta average: %f" % delta_hist.average()
                print "delta histogram:"
                delta_hist.print_top(10)


    exit(0)


    print("Partition histogram:")
    partition_histogram.print_top(n=100)
    print("Counts by extension:")
    extension_len_histogram.print_top(n=100)
    print("Fragments by extension:")
    extension_fragments_histogram.print_top(n=100)

    exit(0)
    for fstype in fstypes:
        if fstype=='(unrecognized)': continue
        print fstype,"Partitions:"

        def isfstype(x): return x.fstype==fstype
        these_parts = filter(isfstype,all_parts)
        these_files = []
        for part in these_parts:
            these_files.extend(part.files)
        print fragmentation_table(these_files)

    
    exit(0)

    sys.exit(0)


    #
    # Typeset information about file extensions
    #
    hist_exts = histogram2d()
    hist_exts.topn = 20
    for i in all_files:
        if i.size>0 and i.fragments>0: hist_exts.add(i.ext(),i.size)
    tab = table()
    tab.header     = "File extension popularity and average size (suppressing 0-len files)"
    tab.col_headings = ['Ext','Count','Average Size','Max','Std Dev']
    tab.omit_row = [[0,'']]
    hist_exts.statcol = ['iaverage','maxx','istddev']
    print hist_exts.typeset(t=tab)

    hist_exts = histogram2d()
    hist_exts.topn = 20
    for i in all_files:
        if i.fragments>0: hist_exts.add(i.ext(),i.fragments)
    tab = table()
    tab.header     = "Fragmentation by file extension (suppressing files with 0 fragments)"
    tab.col_headings = ['Ext','Count','Avg Fragments','Max','Std Dev']
    tab.omit_row = [[0,'']]
    hist_exts.statcol = ['average','maxx','stddev']
    print hist_exts.typeset(t=tab)

    print "==========================="


    #
    # Typeset the File Systems on Drives table
    #

    tab = table()
    tab.header     = "File Systems on Drives"
    tab.col_headings = ["FS Type","Drives","MBytes"]
    tab.col_totals = [1,2]
    fstypeh.statcol = 'sumx'
    print fstypeh.typeset(t=tab)

    #
    # Typeset overall fragmentation stats
    #

    print fragmentation_table(all_files)

if(__name__=="__main__"):
    from optparse import OptionParser
    from copy import deepcopy
    global options

    parser = OptionParser()
    parser.usage="%prog [options] file1 [file2...] (files can be XML or image files)"
    (options,args) = parser.parse_args()

    for fn in args:
        process_files(fn)




import Objects
import logging
import os
import xml.etree.ElementTree as ET

_logger = logging.getLogger(os.path.basename(__file__))
logging.basicConfig(level=logging.DEBUG)

br1 = Objects.ByteRun(img_offset=1, len=1)
br2 = Objects.ByteRun(img_offset=2, len=2)
br3 = Objects.ByteRun(img_offset=4, len=3)

dbr = Objects.ByteRuns()
ibr = Objects.ByteRuns()
nbr = Objects.ByteRuns()

dbr.append(br1)
ibr.append(br2)
nbr.append(br3)

dbr.facet = "data"
ibr.facet = "inode"
nbr.facet = "name"

f1 = Objects.FileObject()
f1.data_brs = dbr
f1.inode_brs = ibr
f1.name_brs = nbr

assert f1.data_brs[0].img_offset == 1
assert f1.inode_brs[0].img_offset == 2
assert f1.name_brs[0].img_offset == 4

e1 = f1.to_Element()
#_logger.debug(f1)
#_logger.debug(ET.tostring(e1))

f2 = Objects.FileObject()

f2.populate_from_Element(e1)
#_logger.debug(f2)

assert f2.data_brs[0].img_offset == 1
assert f2.inode_brs[0].img_offset == 2
assert f2.name_brs[0].img_offset == 4



import Objects
import copy
import logging
import os

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(os.path.basename(__file__))

br = Objects.ByteRun()
br.file_offset = 4128
br.len = 133
brs = Objects.ByteRuns()
brs.append(br)

cbrs1 = copy.deepcopy(brs)

_logger.debug("brs  = %r." % brs)
_logger.debug("cbrs1 = %r." % cbrs1)
assert cbrs1 == brs

cbrs1[0].file_offset += 133
_logger.debug("cbrs1 = %r." % cbrs1)
assert cbrs1 != brs

cbrs2 = copy.deepcopy(brs)
cbrs2[0].type = "unknown"
assert cbrs2 != brs





import logging
import os

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(os.path.basename(__file__))

import Objects

co = Objects.CellObject()

_logger.debug("co = %r" % co)
_logger.debug("co.to_regxml() = %r" % co.to_regxml())

co.name_type = "v"

#Test value-type tolerance of data_type: should be null, strs and ints.

co.data_type = None
co.data_type = 0
co.data_type = "REG_NONE"
failed = False
try:
    co.data_type = 0.1
except:
    failed = True
assert failed



#!/usr/bin/python
"""ground truth program.

Inputs: Multiple image files.
Outputs: List of all the files and partial files in the last file.

Process:

1. Read current disk image. Map out all allocated files. Add each fileinfo to the ground truth.
2. Read all of the previous disk images (or XML files). 
   - If the a previously allocated file is present in the current image, add it to the ground truth.
3. Read all of the previous disk images.
   - For all of the sectors not used in the final version, note which sectors of the original files survive.
"""

import dfxml,fiwalk
import sys
import xml.dom.minidom
from xml.dom.minidom import parseString

# http://wiki.python.org/moin/MiniDom

def make_residual(fi=None,image=None,runs=None):
    """Take a regular fileobject XML document
    and move the byte_runs into an <original> section and add a
    <residual> section."""
    fidoc = fi.doc
    dom = parseString("<foo/>")
    newdoc = deepcopy(fidoc)
    original = dom.createElement("original")
    newdoc.appendChild(original)
    original.appendChild(newdoc.getElementsByTagName("byte_runs")[0])

    # Create the residual
    residual = dom.createElement("residual")
    newdoc.appendChild(residual)

    # Create a new byte_runs
    new_byte_runs = dom.createElement("byte_runs")
    residual.appendChild(new_byte_runs)

    if image:
        x = dom.createElement("image")
        txt = dom.createTextNode(image)
        x.appendChild(txt)
        original.appendChild(x)

    for run in runs:
        x = dom.createElement("run")
        x.setAttribute("img_offset",run.img_offset)
        x.setAttribute("len",run.bytes)
        new_byte_runs.appendChild(x)
    return newdoc
    
def annotate(xml,tag,value):
    from xml.dom.minidom import parseString
    dom = parseString("<foo/>")
    x = dom.createElement(tag)
    txt = dom.createTextNode(value)
    x.appendChild(txt)
    xml.appendChild(x)

def add_reference(doc,image="",reason=""):
    d2 = parseString("<reference><image>%s</image><reason>%s</reason></reference>"
                     % (image,reason))
    doc.appendChild(d2.getElementsByTagName("reference")[0])

def print_ground_truth_fi(fi,out=sys.stdout):
    import os.path
    reference = fi.doc.getElementsByTagName("reference")[0]
    image = reference.getElementsByTagName("image")[0].firstChild.wholeText
    try:
        out.write("Original Filename: %s  in: %s\nSHA 1: %s\n" %
                  (fi.filename(),os.path.basename(image),fi.sha1()))
    except KeyError:
        print "***",fi
        print "***",fi.filename()
        print "***",os.path.basename(image)
    desc = "Location: "
    sector_size = 512
    for run in fi.byte_runs():
        eblurb = ""
        if run.extra_bytes()>0:
            eblurb = "and %3d bytes " % run.extra_bytes()
        out.write("%10s @ sector %8d ; %5d %4d-byte sectors %14s (%7d bytes total)\n"
                  % (desc,run.start_sector(),run.sector_count(),run.sector_size,eblurb,run.bytes))
        desc = ""
    out.write("\n")

def print_ground_truth_report(doc,out=sys.stdout):
    def sort_by_runs(a,b):
        if len(a.byte_runs())==0: return -1
        if len(b.byte_runs())==0: return 1
        if a.byte_runs()[0].img_offset < b.byte_runs()[0].img_offset: return -1
        if a.byte_runs()[0].img_offset > b.byte_runs()[0].img_offset: return 1
        return 0

    filist = [dfxml.fileobject_dom(x)
              for x in doc.getElementsByTagName("fileobject")]
    for fi in sorted(filist,sort_by_runs):
        print_ground_truth_fi(fi,out=out)


def sector_from_file(imagefile,sector_number,sectorsize = 512):
    imagefile.seek(sector_number * sectorsize)
    return imagefile.read(sectorsize)

if __name__=="__main__":
    from optparse import OptionParser
    from copy import deepcopy

    parser = OptionParser()
    parser.usage = '%prog [options] [mapfile1.iso mapfile2.iso ...] masterfile.iso'
    parser.add_option("-x","--xml",help="specify output file for XML",dest="xmlfilename")
    parser.add_option("-r","--report",help="specify output file for the report",dest="reportfilename")
    parser.add_option("-d","--debug",help="debug")
    (options,args) = parser.parse_args()

    if len(args)<1:
        parser.print_help()
        sys.exit(1)

    masterfn = args[-1]
    refs     = args[:-1]
    master_imagefile = open(masterfn,"r")
    db     = dfxml.extentdb(sectorsize=512)

    (doc,fileobjects) = fiwalk.fileobjects_using_dom(imagefile=master_imagefile,
                                                     flags=fiwalk.ALLOC_ONLY)

    groundtruth = parseString("<groundtruth/>")

    # First, add all relevant metadata elements from the master file's
    # XML to the to the groundtruth file's XML.
    for node in doc.childNodes[0].childNodes:
        if node.nodeType==node.ELEMENT_NODE and \
               node.nodeName not in ["volume","fileobject"]:
            groundtruth.childNodes[0].appendChild(node.cloneNode(node))

    # Next, add the file object elements
    for fi in fileobjects:
        if options.debug: print "adding ",fi
        newdoc = fi.doc.cloneNode(fi)
        add_reference(newdoc,image=masterfn,reason='resident file')
        groundtruth.childNodes[0].appendChild(newdoc)
        db.add_runs(fi.byte_runs())
        
    # For each map file, see if any of the allocated files are
    # in the ground truth file but not previously discovered

    for ref in refs:
        if options.debug: print "check files in ",ref
        (d2,fobj2) = fiwalk.fileobjects_using_dom(imagefile=open(ref,"r"),
                                                  flags=fiwalk.ALLOC_ONLY)
        for fi in fobj2:
            runs = fi.byte_runs()
            if not db.intersects_runs(runs):
                db.add_runs(runs)
                newdoc = fi.doc.cloneNode(fi)
                add_reference(newdoc,image=ref,reason='residual file')
                groundtruth.childNodes[0].appendChild(newdoc)


    # Now, for each file, get a list of the sectors that are in unallocated
    # space and report which of them (if any) are in the final file.

    for ref in refs:
        if options.debug: print "check residual data in ",ref
        ref_imagefile = open(ref,"r")
        (d2,fobj2) = fiwalk.fileobjects_using_dom(imagefile=ref_imagefile,
                                                  flags=fiwalk.ALLOC_ONLY)
        for fi in fobj2:
            for run in fi.byte_runs():
                sectors_to_check =  db.sectors_not_in_db(run)
                # For each sector to check, see if the value in the current image file
                # is the same as in the report imagefile

                def check_sector(n):
                    return sector_from_file(ref_imagefile,n) \
                           == sector_from_file(master_imagefile,n)
                sectors_that_match = filter(check_sector,sectors_to_check)
                if sectors_that_match:
                    if options.debug:
                        print(fi.filename(),
                              "run sectors:", db.sectors_for_bytes(run_len),
                              "total sectors: ",len(sectors_to_check),
                              "matching:",len(sectors_that_match))
                    runs = db.runs_for_sectors(sectors_that_match)
                    if options.debug: print("runs:",runs)
                    db.add_sectors(sectors_that_match)
                    residual_doc = make_residual(fi=fi, image=ref, runs=runs)
                    residual = residual_doc.childNodes[0]
                    groundtruth.childNodes[0].appendChild(residual)

    try:
        xmlfile = open(options.xmlfilename,"w")
    except TypeError:
        xmlfile = sys.stdout
        print("Here is the XML:")

    xmlfile.write(groundtruth.toxml())

    try:
        reportfile = open(options.reportfilename,"w")
    except TypeError:
        reportfile = sys.stdout
        print("\n\nHere is the report:")

    print_ground_truth_report(groundtruth,out=reportfile)
    sys.exit(0)





#!/usr/bin/env python
"""rdifference.py

Generates a report about what's different between two Windows Registry hives.

Process:

1. For each regxml file, read all of the fileobject objects:
    - create new maps
    - Note when things change.
    - Delete each file in the old map as it is processed.
2. Report cells left in map; those are the cells that were deleted!
3. Replace the old maps with the new maps
"""

#AJN This script does not call out duplicate paths, but they are reported.

import sys,fiwalk,dfxml,time
if sys.version_info < (3,1):
    raise RuntimeError("rdifference.py requires Python 3.1 or above")

def ptime(t):
    """Print the time in the requested format. T is a dfxml time value"""
    global options
    if t is None:
        return None
    elif "options" in globals() and options.timestamp:
        return str(t.timestamp())
    else:
        return str(t.iso8601())

def dprint(x):
    global options
    if "options" in globals() and options.debug: print(x)

def header():
    if options.html:
        print("""<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN http://www.w3.org/TR/html4/loose.dtd">
<html>
<body>
<style>
body  { font-family: Sans-serif;}
.sha1 { font-family: monospace; font-size: small;}
.filesize { padding-left: 15px; padding-right: 15px; text-align: right;}
</style>
""")

def h1(title):
    global options
    if "options" in globals() and options.html:
        print("<h1>%s</h1>" % title)
        return
    print("\n\n%s\n" % title)

def h2(title):
    global options
    if "options" in globals() and options.html:
        print("<h2>%s</h2>" % title)
        return
    print("\n\n%s\n%s" % (title,"="*len(title)))


def table(rows,styles=None,break_on_change=False):
    import sys
    global options
    def alldigits(x):
        if type(x)!=str: return False
        for ch in x:
            if ch.isdigit()==False: return False
        return True

    def fmt(x):
        if x==None: return ""
        if type(x)==int or alldigits(x):
            return "{0:>12}".format(x)
        return str(x)
            
    if "options" in globals() and options.html:
        print("<table>")
        for row in rows:
            print("<tr>")
            if not styles:
                styles = [""]*len(rows)
            for (col,style) in zip(row,styles):
                sys.stdout.write("<td class='%s'>%s</td>" % (style,col))
            print("<tr>")
        print("</table>")
        return
    lastRowCol0 = None
    for row in rows:
        # Insert a blank line if this row[0] is not the same as last row[0]
        if row[0]!=lastRowCol0 and break_on_change:
            sys.stdout.write("\n")
            lastRowCol0 = row[0]
        # Write the row.
        # This won't generate a unicode encoding error because the strings are valid unicode.
        sys.stdout.write("\t".join([fmt(col) for col in row]))
        sys.stdout.write("\n")

#
# This program keeps track of the current and previous hivestate in a single
# object called "HiveState". Another way to do that would have been to have
# the instance built from the XML file and then have another function that compares
# them.
#        

class HiveState:
    global options

    def __init__(self,notimeline=False):
        self.new_cnames = dict() # maps from cell full path -> cell
        self.notimeline = notimeline
        self.next()
        
    def next(self):
        """Called when the next image is processed."""
        self.cnames = self.new_cnames
        self.new_cnames = dict()
        self.new_files          = set()     # set of file objects
        self.changed_content    = set()     # set of (oldfile,newfile) file objects
        self.changed_properties = set()     # list of (oldfile,newfile) file objects
        if self.notimeline:
            self.timeline = None
        else:
            self.timeline = set()

    def process_cell(self,cell):
        dprint("processing %s" % str(cell))
        
        # See if the filename changed its hash code
        changed = False


        # Remember the file for the next generation
        self.new_cnames[cell.full_path()] = cell
        #new_inodes from idifference translates to ... well, there is no 'inode' in the Registry. There's hive identifier and full path. Values are small and thus more likely to appear in multiple places.  Skip translating

        # See if this filename changed or was resized
        ocell = self.cnames.get(cell.full_path(),None)
        if ocell:
            dprint("   found ocell: " + cell.full_path())
            if ocell.sha1()!=cell.sha1():
                dprint("      >>> sha1 changed")
                self.changed_content.add((ocell,cell))

            if ocell.mtime() != cell.mtime():
                dprint("      >>> Mtime changed")
                self.changed_properties.add((ocell,cell))
            elif ocell.type() != cell.type():
                dprint("      >>> Cell content type changed")
                self.changed_properties.add((ocell,cell))
            elif type(ocell) != type(cell):
                dprint("      >>> Cell structural type changed")
                self.changed_properties.add((ocell,cell))
            elif ( (ocell.parent_key and ocell.parent_key.mtime()) or None ) != ( (cell.parent_key and cell.parent_key.mtime()) or None ):
                dprint("      >>> Parent mtimes changed")
                self.changed_properties.add((ocell,cell))

        # If a new file, note that (and optionally add to the timeline)
        if not ocell:
            self.new_files.add(cell)
            if self.timeline:
                modify_time = cell.mtime()
                self.timeline.add((modify_time,cell.full_path(),"modified"))

        # Delete files we have seen (so we can find out the files that were deleted)
        if cell.full_path() in self.cnames:
            del self.cnames[cell.full_path()]

    def process(self,fname):
        self.current_fname = fname
        if fname.endswith(".regxml"):
            reader = dfxml.read_regxml(xmlfile=open(fname,'rb'), callback=self.process_cell)

    def print_cells(self,title,cells):
        h2(title)
        def cdate(cell):
            try:
                return str(ptime(cell.mtime()))
            except TypeError:
                return "n/a"
        res = [(cdate(cell),cell.full_path()) for cell in cells]
        if res:
            table(sorted(res))

    def print_cell2(self,title,cell2s):
        def prtime(t):
            return "%d (%s)" % (t,ptime(t))

        h2(title)
        res = set()
        for(ocell,cell) in cell2s:
            if ocell.sha1() != cell.sha1():
                res.add((ocell.full_path(),"SHA1 changed",ocell.sha1(),"->",cell.sha1()))
                if self.timeline: self.timeline.add((cell.mtime(),cell.full_path(),"SHA1 changed",ocell.sha1(),"->",cell.sha1()))
            if ocell.mtime() != cell.mtime():
                res.add((ocell.full_path(),"mtime changed",ptime(ocell.mtime()),"->",ptime(cell.mtime())))
                if self.timeline: self.timeline.add((cell.mtime(),cell.full_path(),"mtime changed",prtime(ocell.mtime()),"->",prtime(cell.mtime())))
            if ocell.type() != cell.type():
                res.add((ocell.full_path(),"cell type changed",ocell.type(),"->",cell.type()))
                if self.timeline: self.timeline.add((cell.mtime(),cell.full_path(),"cell type changed",prtime(ocell.mtime()),"->",prtime(cell.mtime())))

        if res:
            table(sorted(res),break_on_change=True)

    def print_timeline(self):
        prt = []

        # Make the dates printable
        for line in sorted(self.timeline):
            prt.append([ptime(line[0])]+list(line[1:]))
        h2("Timeline")
        table(prt)

    def report(self):
        header()
        h1("RegXML file:"+self.current_fname)
        self.print_cells("New Files:",self.new_files)
        self.print_cells("Deleted Files:",self.cnames.values())
        self.print_cell2("Files with modified content:",self.changed_content)
        self.print_cell2("Files with changed file properties:",self.changed_properties)
        if self.timeline: self.print_timeline()

    def output_archive(self,tarname=None,zipname=None):
        """Write the changed and/or new files to a tarfile or a ZIP file. """
        import zipfile, tarfile, StringIO, datetime

        tfile = None
        zfile = None

        to_archive = self.new_files.copy()
        to_archive = to_archive.union(set([val[1] for val in self.changed_content]))
        to_archive = to_archive.union(set([val[1] for val in self.changed_properties]))

        if len(to_archive)==0:
            print("No archive created, as no allocated files created or modified")
            return

        if tarname:
            print(">>> Creating tar file: %s" % tarname)
            tfile = tarfile.TarFile(tarname,mode="w")

        if zipname:
            print(">>> Creating zip file: %s" % zipname)
            zfile = zipfile.ZipFile(zipname,mode="w",allowZip64=True)

        files_written=set()
        content_error_log = []
        for fi in to_archive:
            filename = fi.filename()
            fncount = 1
            while filename in files_written:
                filename = "%s.%d" % (fi.filename(),fnperm)
                fncount+= 1
            contents = None
            try:
                contents = fi.contents(imagefile)
            except ValueError as ve:
                if ve.message.startswith("icat error"):
                    #Some files cannot be recovered, even from images that do not seem corrupted; log the icat command that failed.
                    content_error_log.append(ve.message)
                else:
                    #This is a more interesting error, so have process die to report immediately.
                    raise
            if contents:
                if tfile:
                    info = tarfile.TarInfo(name=filename)
                    info.mtime = fi.mtime()
                    info.atime = fi.atime()
                    info.ctime = fi.ctime()
                    info.uid   = fi.uid()
                    info.gid   = fi.gid()
                    info.size  = fi.filesize()
                    # addfile requires a 'file', so let's make one
                    string = StringIO.StringIO()
                    string.write(contents)
                    string.seek(0)
                    tfile.addfile(tarinfo=info,fileobj=string)
                if zfile:
                    mtimestamp = fi.mtime().timestamp()
                    info = zipfile.ZipInfo(filename)
                    if mtimestamp:
                        #mtime might be null
                        info.date_time = datetime.datetime.fromtimestamp(mtimestamp).utctimetuple()
                    info.internal_attr = 1
                    info.external_attr = 2175008768 # specifies mode 0644
                    zfile.writestr(info,contents)
        if tfile: tfile.close()
        if zfile: zfile.close()
        if len(content_error_log) > 0:
            sys.stderr.write("Errors retrieving file contents:\n")
            sys.stderr.write("\n".join(content_error_log))
            sys.stderr.write("\n")

if __name__=="__main__":
    from optparse import OptionParser
    from copy import deepcopy

    parser = OptionParser()
    parser.usage = '%prog [options] file1 file2 [file3...]  (files can be xml or image files)'
    parser.add_option("-x","--xml",help="specify output file for XML",dest="xmlfilename")
    parser.add_option("--html",help="specify output in HTML",action="store_true")
    parser.add_option("-n","--notimeline",help="do not generate a timeline",action="store_true")
    parser.add_option("-d","--debug",help="debug",action='store_true')
    parser.add_option("-T","--tararchive",help="create tar archive file of new/changed files",dest="tarfile")
    parser.add_option("-Z","--zipfile",help="create ZIP64 archive file of new/changed files",dest="zipfile")
    parser.add_option("--timestamp",help="output all times in Unix timestamp format; otherwise use ISO 8601",action="store_true")

    (options,args) = parser.parse_args()

    if len(args)<1:
        parser.print_help()
        sys.exit(1)

    s = HiveState(notimeline=options.notimeline)
    for infile in args:
        print(">>> Reading %s" % infile)
        s.process(infile)
        if infile!=args[0]:
            # Not the first file. Report and optionally archive
            if options.tarfile or options.zipfile:
                s.output_archive(tarname=options.tarfile,zipname=options.zipfile)
            s.report()
        s.next()



import sys, fiwalk, os.path
from optparse import OptionParser
from sys import stdout
 
def demo_dfxml_time_bug(filename):
    parser = OptionParser()
    parser.usage = '%prog% [options] xmlfile '
    (options,args) = parser.parse_args()
    for fi in fiwalk.fileobjects_using_sax(xmlfile=open(filename,"rb")):
        fsize = fi.filesize()
        try:
            mt = fi.mtime()
            print('Type of mt:',type(mt))
            print('Normal mtime:')
            print(mt)
        except KeyboardInterrupt:
            raise
        except:
            raise
            print('Abnormal mtime for file with size',fsize)
 
if __name__=="__main__":
    filename = sys.argv[1]
    demo_dfxml_time_bug(filename)



#!/usr/bin/python
"""Draw a quick hisogram of the timestamps on the hard drive"""

import matplotlib
matplotlib.use('agg.pdf')


import fiwalk
import datetime
from matplotlib.dates import MonthLocator, WeekdayLocator, DateFormatter
from matplotlib.dates import MONDAY,SATURDAY
import time
from pylab import *


def get_dates_and_counts(times):
    from datetime import date
    data = {}
    for t in times:
        gm = time.gmtime(t)
        d = date(gm[0],gm[1],gm[2])
        data[d] = data.get(d,0)+1

    # Create a list of key,val items so you can sort by date
    dates_and_counts = [ (date,count) for date,count in data.items()]
    dates_and_counts = sorted(dates_and_counts)
    return dates_and_counts

def version1(times):    
    import pylab
    pylab.grid()
    pylab.hist(times,100)
    pylab.show()


def version2(times):
    # see http://mail.python.org/pipermail/python-list/2003-November/236559.html
    # http://www.gossamer-threads.com/lists/python/python/665014
    from matplotlib.pylab import plot, show, title, xlabel, ylabel, gca, bar, savefig, plot_date
    
    dates_and_counts = get_dates_and_counts(times)
    dates, counts = zip(*dates_and_counts)
    # bar(dates,counts)
    plot_date(dates,counts)
    xlabel("Date")
    ylabel("count")
    show()

def version3(times):
    import datetime
    import numpy as np
    import matplotlib
    import matplotlib.pyplot as pyplot
    import matplotlib.dates as mdates
    import matplotlib.mlab as mlab

    dates_and_counts = get_dates_and_counts(times)
    dates, counts = zip(*dates_and_counts)

    years    = mdates.YearLocator()   # every year
    months   = mdates.MonthLocator()  # every month
    yearsFmt = mdates.DateFormatter('%Y')

    fig = pyplot.figure()
    ax = fig.add_subplot(111)
    ax.bar(dates,counts)

    ax.set_ylabel('file count')
    ax.set_xlabel('file modification time (mtime)')

    #ax.set_yscale('log')

    # Format the ticks

    ax.xaxis.set_major_locator(years)
    ax.xaxis.set_major_formatter(yearsFmt)
    #ax.xaxis.set_minor_locator(months)

    datemin = datetime.date(min(dates).year, 1, 1)
    datemax = datetime.date(max(dates).year, 1, 1)
    ax.set_xlim(datemin, datemax)
    ax.set_ylim(0,max(counts))

    # format the coords message box
    def price(x): return '$%1.2f'%x
    ax.format_xdata = mdates.DateFormatter('%Y-%m-%d')
    ax.format_ydata = price
    ax.grid(True)

    # rotates and right aligns the x labels, and moves the bottom of the
    # axes up to make room for them
    fig.autofmt_xdate()
    plt.savefig("hist.pdf",format='pdf')
                
    print("dates:",dates)
    print("num dates:",len(dates))

    
    
if __name__=="__main__":
    import sys
    from optparse import OptionParser
    from sys import stdout

    parser = OptionParser()
    parser.usage = '%prog [options] xmlfile '
    (options,args) = parser.parse_args()

    import time
    times = []
    for fi in fiwalk.fileobjects_using_sax(xmlfile=open(args[0])):
        try:
            times.append(fi.mtime())
        except KeyError:
            pass

    version3(times)

    




import Objects
import logging
import os

_logger = logging.getLogger(os.path.basename(__file__))
logging.basicConfig(level=logging.DEBUG)

br = Objects.ByteRun()
_logger.debug("br = %r." % br)

assert br.fill is None

br.fill = b'\x00'
_logger.debug("br.fill = %r." % br.fill)
assert br.fill == b'\x00'

#Catch current implementation decision.
multibyte_failed = None
try:
    br.fill = b'\x00\x01'
except NotImplementedError as e:
    multibyte_failed = True
assert multibyte_failed

br.fill = 0
_logger.debug("br.fill = %r." % br.fill)
assert br.fill == b'\x00'

br.fill = "0"
_logger.debug("br.fill = %r." % br.fill)
assert br.fill == b'\x00'

br.fill = 1
_logger.debug("br.fill = %r." % br.fill)
assert br.fill == b'\x01'



#!/usr/bin/python
"""Redact an image file using a ruleset... 

Image Redaction Project.

This program redacts disk image files. 

inputs:
	* The disk image file
	* A set of rules that describe what to redact, and how to redact it.

Rule File format:

The readaction command file consists of commands. 
Each command has an "condition" and an "action"

  [condition] [action]

Conditions:
  FILENAME <afilename> - a file with the given name
  FILEPAT  <a file pattern> - any file with a given pattern
  DIRNAME  <a directory> - any file in the directory 
  MD5 <a md5> - any file with the given md5
  SHA1 <a sha1> - any file with the given sha1 
  CONTAINS <a string> - any file that contains <a string>

Actions:
  SCRUB MATCH - Scrubs the pattern where it occures
  SCRUB SECTOR - Scrubs the block where the patern occures
  SCRUB FILE - Scrubs the file in which the pattern occures

Actions:
   FILL 0x44	- overwrite by filling with character 0x44 ('D')
   ENCRYPT      - encrypts the data
   FUZZ		- fuz the binary, but not the strings

Examples:

Example file:
===============

MD5 3482347345345 SCRUB FILE
MATCH simsong@acm.org SCRUB FILE
MATCH foobar SCRUB BLOCK
================================================================
Other actions in file:

KEY 12342343  (an encryption key)
"""


import xml.parsers.expat
import hashlib
import os.path
import fiwalk
import re

################################################################
def convert_fileglob_to_re(fileglob):
    regex = fileglob.replace(".","[.]").replace("*",".*").replace("?",".?")
    return re.compile(regex)

class redact_rule:
    """ Instances of this class are objects that can decide whether or not to redact."""
    def __init__(self,line):
        self.line = line
        self.complete = True               # by default, redacts everything
    def should_redact(self,fileobject):
        """Returns True if this fileobject should be redacted"""
        raise ValueError("redact method of redact_rule super class should not be called")
    def __str__(self):
        return "action<"+self.line+">"
    def runs_to_redact(self,fi):
        """Returns the byte_runs of the source which match the rule.
        By default this is the entire object."""
        return fi.byte_runs()

    
class redact_rule_md5(redact_rule):
    """ redact if the MD5 matches"""
    def __init__(self,line,val):
        redact_rule.__init__(self,line)
        self.md5val = val.lower()
    def should_redact(self,fi):
        return self.md5val == fi.tag('md5')
    
class redact_rule_sha1(redact_rule):
    """ redact if the SHA1 matches"""
    def __init__(self,line,val):
        redact_rule.__init__(self,line)
        self.sha1val = val.lower()
    def should_redact(self,fi):
        return self.sha1val == fi.tag('sha1')
    
class redact_rule_filepat(redact_rule):
    def __init__(self,line,filepat):
        import re
        redact_rule.__init__(self,line)
        # convert fileglobbing to regular expression
        self.filepat_re = convert_fileglob_to_re(filepat)
        print("adding rule to redact path "+self.filepat_re.pattern)
    def should_redact(self,fileobject):
        return self.filepat_re.search(fileobject.filename())

class redact_rule_filename(redact_rule):
    def __init__(self,line,filename):
        redact_rule.__init__(self,line)
        self.filename = filename
        print("adding rule to redact filename "+self.filename)
    def should_redact(self,fileobject):
        was = os.path.sep
        os.path.sep = '/'                       # Force Unix filename conventions
        ret = self.filename == os.path.basename(fileobject.filename())
        os.path.sep = was
        return ret

class redact_rule_dirname(redact_rule):
    def __init__(self,line,dirname):
        redact_rule.__init__(self,line)
        self.dirname = dirname
    def should_redact(self,fileobject):
        was = os.path.sep
        os.path.sep = '/'                      # Force Unix filename conventions
        ret = self.dirname == os.path.dirname(fileobject.filename())
        os.path.sep = was
        return ret

class redact_rule_contains(redact_rule):
    def __init__(self,line,text):
        redact_rule.__init__(self,line)
        self.text = text
    def should_redact(self,fileobject):
        return self.text in fileobject.contents()

class redact_rule_string(redact_rule):
    def __init__(self,line,text):
        redact_rule.__init__(self,line)
        self.text = text
        self.complete = False           # doesn't redact the entire file

    def should_redact(self,fileobject):
        return self.text in fileobject.contents()
    def runs_to_redact(self,fi):
        """Overridden to return the byte runs of just the given text"""
        ret = []
        tlen = len(self.text)
        for run in fi.byte_runs():
            (file_offset,run_len,img_offset) = run

            run_content = fi.content_for_run(run)
            offset = 0
            # Now find all the places inside "run"
            # where the text "self.text" appears
            print("looking for '{}' in '{}'".format(self.text,run))
            while offset>=0:
                offset = run.find(self.text,offset)
                if offset>=0:
                    ret.append((file_offset+offset,tlen,img_offset+offset))
                    offset += 1         # 
        return ret

"""Not actually a redact rule, but rather a rule for global ignores"""
class ignore_rule():
    def __init__(self):
        self.ignore_patterns = []
    def ignore(self,ignore):
        """Ignores specified files based on a regex"""
        self.ignore_patterns.append(re.compile(convert_fileglob_to_re(ignore)))
        return self
    def should_ignore(self, fi):
        for ig in self.ignore_patterns:
            if ig.search(fi.filename()): 
                return True
        return False

################################################################
class redact_action():
    """Instances of this class are objects that specify how a redaction shoudl be done."""
    def redact(self,rule,fileobject,rc):
        """Performs the redaction"""
        raise ValueError,"redact method of redact_action super class should not be called"

class redact_action_fill(redact_action):
    """ Perform redaction by filling"""
    def __init__(self,val):
        self.fillvalue = val
    def redact(self,rule,fi,rc):
        for run in rule.runs_to_redact(fi):
            print("   Current run %s " % run)
            rc.imagefile.seek(run.img_offset)
            runlen = run.len
            print "\tFile info - \n\t\tname: %s  \n\t\tclosed: %s \n\t\tposition: %d \n\t\tmode: %s" % \
            (rc.imagefile.name, rc.imagefile.closed, rc.imagefile.tell(), rc.imagefile.mode)
            print("   Filling at offset {}, {} bytes with pattern {}".format(run.img_offset,runlen,hex(self.fillvalue)))
            if rc.commit:
                rc.imagefile.seek(run.img_offset)
                rc.imagefile.write(chr(self.fillvalue) * run.len)
                print("   >>COMMIT\n")
            
class redact_action_encrypt(redact_action):
    """ Perform redaction by encrypting"""
    def redact(self,rule,fileobject,rc):
        for run in rule.runs_to_redact(fileobject):
            print("   encrypting at offset {}, {} bytes with cipher".format(run.img_offset,run.bytes))
            raise ValueError,"Whoops; Didn't write this yet"
            
class redact_action_fuzz(redact_action):
    """ Perform redaction by fuzzing x86 instructions """
    def redact(self,rule,fileobject,rc):
        '''
        The net effect of this function is that bytes 127-255 are "fuzzed" over 
        the range of 159-191, with each series of four bytes
        (e.g. 128-131) to one byte value (e.g. 160).
        '''
        def fuzz(ch):
            o = ord(ch)
            if(o<127):
                r = ch
            else:
                r = chr(((o>>2)+128)%256)
            return r
        print "Redacting with FUZZ: ",fileobject
        for run in rule.runs_to_redact(fileobject):
            try:
                print "   Fuzzing at offset: %d, can fuzz up to %d bytes " % (run.img_offset,run.len)
                rc.imagefile.seek(run.img_offset)

                # Previously redacted only first 10 bytes, now redacts entire sequence
                #first_ten_bytes = rc.imagefile.read(10)
                run_bytes = rc.imagefile.read(run.len)

                print "\tFile info - \n\t\tname: %s  \n\t\tclosed: %s \n\t\tposition: %d \n\t\tmode: %s" % \
                    (rc.imagefile.name, rc.imagefile.closed, rc.imagefile.tell(), rc.imagefile.mode)
                print "    Fuzzing %d bytes - should be %d" % (len(run_bytes), run.len)
                newbytes = "".join([fuzz(x) for x in run_bytes])
                #debug
                print "new: %i old: %i" % (len(newbytes), run.len)
                assert(len(newbytes)==run.len)
                if rc.commit:
                    rc.imagefile.seek(run.img_offset)
                    rc.imagefile.write(newbytes)
                    print "\n   >>COMMIT"
            except AttributeError:
                print "!AttributeError: no byte run?"

            
################################################################
class RedactConfig:
    """Class to read and parse a redaction config file"""
    def __init__(self,fn):
        self.cmds = []
        self.commit = False
        self.filename = None
        self.xmlfile = None
        self.ignore_rule = ignore_rule()
        for line in open(fn,"r"):
            if line[0] in '#;': continue       # comment line
            line = line.strip()
            if line=="": continue
            atoms = line.split(" ")
            while "" in atoms: atoms.remove("") # take care of extra spaces
            cmd = atoms[0].lower()
            rule = None
            action = None

            # First look for simple commands

            if cmd=='key':
                self.key = atoms[1]
                continue

            if cmd=="commit":
                self.commit = True
                continue

            if cmd=="imagefile":
                self.imagefile = open(atoms[1],"r+b")
                continue

            if cmd=="xmlfile":
                self.xmlfile = open(atoms[1],"r")
                continue

            if cmd=='ignore':
                self.ignore_rule.ignore(atoms[1])
                continue

            # Now look for commands that are rules

            if cmd=='md5':
                rule = redact_rule_md5(line,atoms[1])
            if cmd=='sha1':
                rule = redact_rule_sha1(line,atoms[1])
            if cmd=='filename':
                rule = redact_rule_filename(line,atoms[1])
            if cmd=='filepat':
                rule = redact_rule_filepat(line,atoms[1])
            if cmd=='contains':
                rule = redact_rule_contains(line,atoms[1])
            if cmd=='string':
                rule = redact_rule_string(line,atoms[1])

            if rule:
                if atoms[2].lower()=='fill':
                    action = redact_action_fill(eval(atoms[3]))
                if atoms[2].lower()=='encrypt':
                    action = redact_action_encrypt()
                if atoms[2].lower()=='fuzz':
                    action = redact_action_fuzz()


            if not rule or not action:
                print "atoms:",atoms
                print "rule:",rule
                print "action:",action
                raise ValueError,"Cannot parse: '%s'" % line
            self.cmds.append((rule,action))

    def need_md5(self):
        for (rule,action) in self.cmds:
            if rule.__class__==redact_rule_md5: return True
        return False

    def need_sha1(self):
        for (rule,action) in self.cmds:
            if rule.__class__==redact_rule_sha1: return True
        return False

    def fiwalk_opts(self):
        "Returns the options that fiwalk needs given the redaction requested."
        opts = "-x"
        if self.need_sha1(): opts = opts+"1"
        if self.need_md5():  opts = opts+"m"
        return opts

    def process_file(self,fileinfo):
        for (rule,action) in self.cmds:
            if rule.should_redact(fileinfo):
                print "Processing file: %s" % fileinfo.filename()

                if self.ignore_rule.should_ignore(fileinfo):
                    print "(Ignoring %s)" % fileinfo.filename()
                    return

                print ""
                print "Redacting ",fileinfo.filename()
                print "Reason:",str(rule)
                print "Action:",action
                action.redact(rule,fileinfo,self)
                if rule.complete:
                    return                  # only need to redact once!

    def close_files(self):
        if self.imagefile and self.imagefile.closed == False:
            print "Closing file: %s" % self.imagefile.name
            self.imagefile.close()
        if self.xmlfile and self.xmlfile.closed == False:
            print "Closing file: %s" % self.xmlfile.name
            self.xmlfile.close()

if __name__=="__main__":
    import sys,time
    from optparse import OptionParser
    from subprocess import Popen,PIPE
    global options

    parser = OptionParser()
    parser.usage = "%prog [options] config-file"
    parser.add_option("-d","--debug",help="prints debugging info",dest="debug")
    (options,args) = parser.parse_args()

    t0 = time.time()
    # Read the redaction configuration file
    rc = RedactConfig(args[0])

    if not rc.imagefile:
        print "Error: a filename must be specified in the redaction config file"
        sys.exit(1)

    fiwalk.fiwalk_using_sax(imagefile=rc.imagefile,xmlfile=rc.xmlfile,callback=rc.process_file)
    t1 = time.time()

    rc.close_files()

    print "Time to run: %d seconds" % (t1-t0)



#!/usr/bin/python
#
# dedup  - detect and optionally remove duplicates based on a DFXML file

import os,dfxml,xml

class dedup:
    def __init__(self):
        from collections import defaultdict
        self.seen = defaultdict(list)
        self.files = 0
        self.md5s = 0

    def process(self,fi):
        self.files += 1
        if fi.md5():
            self.seen[fi.md5()].append(fi.filename())
            self.md5s += 1

    def find_dups(self,cb=None):
        for (md5,names) in self.seen.items():
            if cb and len(names)>1:
                cb(names)

    def report(self,func,cb):
        for (md5,names) in self.seen.items():
            if func(names): cb(names)

def process_dups(names):
    print("dups: ",names)

if __name__=="__main__":
    from argparse import ArgumentParser
    global options

    parser = ArgumentParser()
    parser.add_argument("dfxml",type=str)
    parser.add_argument("--verbose",action="store_true")
    parser.add_argument("--prefix",type=str,help="Only output files with the given prefix")
    parser.add_argument("--distinct",action='store_true',help='Report the distinct files')
    parser.add_argument("--dups",action='store_true',help='Report the files that are dups, and give dup count')
    args = parser.parse_args()

    dobj = dedup()

    try:
        dfxml.read_dfxml(open(args.dfxml,'rb'),callback=dobj.process)
    except xml.parsers.expat.ExpatError:
        pass

    print("Total files: {:,}  total MD5s processed: {:,}  Unique MD5s: {:,}".format(dobj.files,dobj.md5s,len(dobj.seen)))

    if args.distinct:
        def report_distinct(names):
            if args.prefix and not names[0].startswith(args.prefix): return
            print("distinct: ",names[0])
        dobj.report(lambda names:len(names)==1,report_distinct)

    if args.dups:
        def report_dups(names):
            for name in names:
                if not args.prefix or name.startswith(args.prefix):
                    print("dups: {} {}".format(name,len(names)))
        dobj.report(lambda names:len(names)>1,report_dups)




#!/usr/bin/env python3

"""
For a disk image or DFXML file, this program produces a cross-tabulation of the allocation state of each file's inode and name.
"""

__version__ = "0.1.1"
#Version 0.2.0:
# * Tabular output in HTML
# * Tabular output in LaTeX

import Objects
import make_differential_dfxml

import collections
import logging
import sys
import xml.etree.ElementTree as ET
import os

_logger = logging.getLogger(os.path.basename(__file__))

def main():
    counter = collections.defaultdict(lambda: 0)
    prev_obj = None
    for (event, obj) in Objects.iterparse(args.input_image):
        if isinstance(obj, Objects.FileObject):
            if args.ignore_virtual_files and make_differential_dfxml.ignorable_name(obj.filename):
                continue
            counter[(obj.alloc_inode, obj.alloc_name)] += 1

            #Inspect weird data
            if args.debug and obj.alloc_inode is None and obj.alloc_name is None:
                _logger.debug("Encountered a file with all-null allocation.")
                _logger.debug("Event: %r." % event)
                _logger.debug("Previous object: %s." % ET.tostring(prev_obj.to_Element()))
                _logger.debug("Current  object: %s." % ET.tostring(obj.to_Element()))
        prev_obj = obj
    print(repr(counter))

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--ignore-virtual-files", action="store_true", help="Use the same file-ignoring rules as make_differential_dfxml.py.")
    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug printing.")
    parser.add_argument("input_image", help="Disk image, or DFXML file.")
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    main()



#!/usr/bin/env python

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties. Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain. NIST assumes no
# responsibility whatsoever for its use by other parties, and makes
# no guarantees, expressed or implied, about its quality,
# reliability, or any other characteristic.
#
# We would appreciate acknowledgement if the software is used.

__version__ = "0.1.0"

import logging

import Objects

def main():
    dobj = Objects.DFXMLObject()
    dobj.diff_file_ignores.add("atime")
    dobj.diff_file_ignores.add("crtime")
    with open(args.out_dfxml, "w") as fh:
        dobj.print_dfxml(fh)

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", action="store_true")
    parser.add_argument("out_dfxml")
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    main()



#!/usr/bin/env python

import dfxml,fiwalk
import zipfile,sys,os,os.path,datetime

if __name__=="__main__":
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-x", "--xml", dest="xmlfilename", help="Already-created DFXML file for imagefile")
    parser.usage = '%prog [options] imagefile zipfile [x1 x2 x3]\nFind files x1, x2, x3 ... in imagefile and write to zipfile'
    (options,args) = parser.parse_args()

    if len(args)<3:
        parser.print_help()
        exit(1)

    imagefilename = args[0]
    xmlfilename = options.xmlfilename
    xmlfh = None
    if xmlfilename != None:
        xmlfh = open(xmlfilename, "r")
    zipfilename = args[1]
    targets = set([fn.lower() for fn in args[2:]])
    zfile = zipfile.ZipFile(zipfilename,"w",allowZip64=True)
    
    def proc(fi):
        basename = os.path.basename(fi.filename()).lower()
        if basename in targets:
            info = zipfile.ZipInfo(fi.filename(),datetime.datetime.fromtimestamp(fi.mtime().timestamp()).utctimetuple())
            info.internal_attr = 1
            info.external_attr = 2175008768 # specifies mode 0644
            zfile.writestr(info,fi.contents())
    fiwalk.fiwalk_using_sax(imagefile=open(imagefilename), xmlfile=xmlfh, callback=proc)
    



#!/usr/bin/env python
"""iexport.py: export the unallocated spaces."""


class Run:
    """Keeps track of a single run"""
    def __init__(self,start,len):
        self.start = start
        self.len = len
        self.end = start+len-1
    def __str__(self):
        return "Run<%d--%d> (len %d)" % (self.start,self.end,self.len)
    def contains(self,b):
        """Returns true if b is inside self."""
        print "%d <= %d <= %d = %s" % (self.start,b,self.end,(self.start <= b <= self.end))
        return self.start <= b <= self.end
    def intersects_run(self,r):
        """Return true if self intersects r.  This may be because r.start is
        inside the run, r.end is inside the run, or self is inside the run."""
        return self.contains(r.start) or self.contains(r.end) or r.contains(self.start)
    def contains_run(self,r):
        """Returns true if self completely contains r"""
        return self.contains(r.start) and self.contains(r.end)
        

class RunDB:
    """The RunDB maintains a list of all the runs in a disk image. The
RunDB is created with a single run that represnts all of the sectors
in the disk image. Runs can then be removed, which causes existing
runs to be split. Finally all of the remaining runs can be removed."""
    def __init__(self,start,len):
        self.runs = [ Run(start,len) ]
    def __str__(self):
        return "RunDB\n" + "\n".join([str(p) for p in self.runs])
    def intersecting_runs(self,r):
        """Return a list of all the Runs that intersect with r.
        This may be because r.start is inside the run, r.end is inside
        the run, because the run completely encloses r, or becuase r completely
        encloses the run."""
        return filter(lambda x:x.intersects_run(r) , self.runs)
    def remove(self,r):
        """Remove run r"""
        for p in self.intersecting_runs(r):
            self.runs.remove(p)

            # if P is completely inside r, just remove it
            if r.contains_run(p):
                continue

            # Split p into before and after r; add the non-zero pieces
            before_len = r.start - p.start
            if before_len>0:
                self.runs.append(Run(p.start,before_len))
            after_len = p.end - r.end
            if after_len>0:
                self.runs.append(Run(r.end,after_len))
                

if __name__=="__main__":
    r1 = Run(0,1000)
    r2 = Run(50,60)
    assert r1.intersects_run(r2)
    assert r2.intersects_run(r1)

    disk = RunDB(0,1000)
    print disk
    disk.remove(Run(50,60))
    disk.remove(Run(0,10))
    disk.remove(Run(40,20))
    print disk

